# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Helper functions for working with network manager.
"""

import collections
import datetime
import logging
import socket
import struct
import time
import uuid

from django.utils.translation import gettext_lazy as _

from plinth.utils import import_from_gi

glib = import_from_gi('GLib', '2.0')
nm = import_from_gi('NM', '1.0')

logger = logging.getLogger(__name__)

_client = None

ZONES = [('external', _('External')), ('internal', _('Internal'))]

CONNECTION_TYPE_NAMES = collections.OrderedDict([
    ('802-3-ethernet', _('Ethernet')),
    ('802-11-wireless', _('Wi-Fi')),
    ('pppoe', _('PPPoE')),
    ('generic', _('Generic')),
])


class ConnectionNotFound(Exception):
    """Network connection with a given name could not be found."""


class DeviceNotFound(Exception):
    """Network device for specified operation could not be found."""


def ipv4_string_to_int(address):
    """Return an integer equivalent of a string contain IPv4 address."""
    return struct.unpack('=I', socket.inet_aton(address))[0]


def ipv4_int_to_string(address_int):
    """Return an string equivalent of a integer IPv4 address."""
    return socket.inet_ntoa(struct.pack('=I', address_int))


def init():
    """Create and keep a network manager client instance."""

    def new_callback(source_object, result, user_data):
        """Called when new() operation is complete."""
        global _client
        _client = nm.Client.new_finish(result)
        logger.info('Created Network manager client')

    logger.info('Creating network manager client')
    nm.Client.new_async(None, new_callback, None)


def get_nm_client():
    """Return a network manager client object."""
    if _client:
        return _client

    raise Exception('Client not yet ready')


def _callback(source_object, result, user_data):
    """Called when an operation is completed."""
    del source_object  # Unused
    del result  # Unused
    del user_data  # Unused


def get_interface_list(device_type):
    """Get a list of network interface available on the system."""
    interfaces = {}
    for device in get_nm_client().get_devices():
        if device.get_device_type() == device_type:
            interfaces[device.get_iface()] = device.get_hw_address()

    return interfaces


def _is_primary(connection):
    """Return whether a connection is primary connection."""
    primary = get_nm_client().get_primary_connection()
    return (primary and primary.get_uuid() == connection.get_uuid())


def get_status_from_connection(connection):
    """Return the current status of a connection."""
    status = collections.defaultdict(dict)

    status['id'] = connection.get_id()
    status['uuid'] = connection.get_uuid()
    status['type'] = connection.get_connection_type()
    status['zone'] = connection.get_setting_connection().get_zone()
    status['dns_over_tls'] = \
        connection.get_setting_connection().get_dns_over_tls().value_nick
    status['interface_name'] = connection.get_interface_name()
    status['primary'] = _is_primary(connection)

    status['ipv4']['method'] = connection.get_setting_ip4_config().get_method()
    status['ipv6']['method'] = connection.get_setting_ip6_config().get_method()

    if status['type'] == '802-11-wireless':
        setting_wireless = connection.get_setting_wireless()
        status['wireless']['ssid'] = setting_wireless.get_ssid().get_data(
        ).decode()
        status['wireless']['mode'] = setting_wireless.get_mode()

    return status


def get_status_from_active_connection(connection):
    """Return the current status of an active connection."""
    status = collections.defaultdict(dict)

    status['state'] = connection.get_state().value_name
    status['ip4']['default'] = connection.get_default()
    status['ip6']['default'] = connection.get_default6()

    return status


def get_status_from_device(device):
    """Return a dictionary with current status of a network device."""
    if not device:
        return None

    status = collections.defaultdict(dict)

    ip4_config = device.get_ip4_config()
    if ip4_config:
        addresses = ip4_config.get_addresses()
        status['ip4']['addresses'] = [{
            'address': address.get_address(),
            'prefix': address.get_prefix()
        } for address in addresses]
        status['ip4']['gateway'] = ip4_config.get_gateway()
        status['ip4']['nameservers'] = ip4_config.get_nameservers()

    ip6_config = device.get_ip6_config()
    if ip6_config:
        addresses = ip6_config.get_addresses()
        status['ip6']['addresses'] = [{
            'address': address.get_address(),
            'prefix': address.get_prefix()
        } for address in addresses]
        status['ip6']['gateway'] = ip6_config.get_gateway()
        status['ip6']['nameservers'] = ip6_config.get_nameservers()

    status['type'] = device.get_type_description()
    status['description'] = device.get_description()
    status['hw_address'] = device.get_hw_address()
    status['interface_name'] = device.get_iface()
    status['state'] = device.get_state().value_nick
    status['state_reason'] = device.get_state_reason().value_nick

    if device.get_device_type() == nm.DeviceType.WIFI:
        status['wireless']['bitrate'] = device.get_bitrate() / 1000
        status['wireless']['mode'] = device.get_mode().value_nick

    if device.get_device_type() == nm.DeviceType.ETHERNET:
        status['ethernet']['speed'] = device.get_speed()
        status['ethernet']['carrier'] = device.get_carrier()

    return status


def get_status_from_wifi_access_point(device, ssid):
    """Return the current status of an access point."""
    status = {}

    if not ssid or not device:
        return status

    for access_point in device.get_access_points():
        if access_point and access_point.get_ssid() and \
           access_point.get_ssid().get_data().decode() == ssid:
            status['strength'] = access_point.get_strength()
            frequency = access_point.get_frequency()
            status['channel'] = _get_wifi_channel_from_frequency(frequency)
            break

    return status


def _get_wifi_channel_from_frequency(frequency):
    """Get the wifi channel form a particular SSID"""
    # TODO: Hard coded list of wifi frequencys and their corresponding
    # channel numbers.  Search for a better solution!  Even 5GHz is
    # not included yet.  Only the plain frequency will show up on 5GHz
    # AP's.
    channel_map = {
        2412: 1,
        2417: 2,
        2422: 3,
        2427: 4,
        2432: 5,
        2437: 6,
        2442: 7,
        2447: 8,
        2452: 9,
        2457: 10,
        2462: 11
    }
    try:
        return channel_map[frequency]
    except KeyError:
        return str(frequency / 1000) + 'GHz'


def get_connection_list():
    """Get a list of active and available connections."""
    client = get_nm_client()
    primary_connection = client.get_primary_connection()

    active_uuids = []
    for connection in client.get_active_connections():
        active_uuids.append(connection.get_uuid())

    connections = []
    for connection in client.get_connections():
        # Display a friendly type name if known.
        connection_type = connection.get_connection_type()
        # Do not show bridge adapter as it is not meant to
        # be modified by the user.
        if connection_type == 'bridge':
            continue

        connection_type_name = CONNECTION_TYPE_NAMES.get(
            connection_type, connection_type)

        settings_connection = connection.get_setting_connection()
        zone = settings_connection.get_zone()

        connection.get_interface_name()

        connections.append({
            'name': connection.get_id(),
            'uuid': connection.get_uuid(),
            'interface_name': connection.get_interface_name(),
            'type': connection_type,
            'type_name': connection_type_name,
            'is_active': connection.get_uuid() in active_uuids,
            'primary':
                (primary_connection
                 and primary_connection.get_uuid() == connection.get_uuid()),
            'zone': zone,
        })
    connections.sort(key=lambda connection: connection['is_active'],
                     reverse=True)
    return connections


def get_connection(connection_uuid):
    """Return connection with matching uuid.

    Raise ConnectionNotFound if a connection with that uuid is not
    found.
    """
    client = get_nm_client()
    connection = client.get_connection_by_uuid(connection_uuid)
    if not connection:
        raise ConnectionNotFound(connection_uuid)

    return connection


def get_active_connection(connection_uuid):
    """Return active connection with matching UUID.

    Raise ConnectionNotFound if a connection with that uuid is not
    found.
    """
    connections = get_nm_client().get_active_connections()
    connections = {
        connection.get_uuid(): connection
        for connection in connections
    }
    try:
        return connections[connection_uuid]
    except KeyError:
        raise ConnectionNotFound(connection_uuid)


def get_connection_by_interface_name(interface_name):
    """Return connection with matching interface.

    Return None if a connection with the interface is not found."""
    client = get_nm_client()
    for connection in client.get_connections():
        if connection.get_interface_name() == interface_name:
            return connection

    return None


def get_device_by_interface_name(interface_name):
    """Return a device by interface name."""
    return get_nm_client().get_device_by_iface(interface_name)


def _update_common_settings(connection, connection_uuid, common):
    """Create/edit basic settings for network manager connections.

    Return newly created connection object if connection is None.
    """
    if not connection:
        connection = nm.SimpleConnection.new()

    settings = connection.get_setting_connection()
    if not settings:
        settings = nm.SettingConnection.new()
        connection.add_setting(settings)

    settings.set_property(nm.SETTING_CONNECTION_UUID, connection_uuid)
    if 'name' in common:
        settings.set_property(nm.SETTING_CONNECTION_ID, common['name'])

    if 'type' in common:
        settings.set_property(nm.SETTING_CONNECTION_TYPE, common['type'])

    if 'interface' in common:
        settings.set_property(nm.SETTING_CONNECTION_INTERFACE_NAME,
                              common['interface'])

    if 'zone' in common:
        settings.set_property(nm.SETTING_CONNECTION_ZONE, common['zone'])

    if 'dns_over_tls' in common:
        values = {
            'default': nm.SettingConnectionDnsOverTls.DEFAULT,
            'no': nm.SettingConnectionDnsOverTls.NO,
            'opportunistic': nm.SettingConnectionDnsOverTls.OPPORTUNISTIC,
            'yes': nm.SettingConnectionDnsOverTls.YES
        }
        settings.set_property(nm.SETTING_CONNECTION_DNS_OVER_TLS,
                              values[common['dns_over_tls']])

    if 'autoconnect' in common:
        settings.set_property(nm.SETTING_CONNECTION_AUTOCONNECT,
                              common['autoconnect'])

    return connection


def _update_ipv4_settings(connection, ipv4):
    """Edit IPv4 settings for network manager connections."""
    settings = nm.SettingIP4Config.new()
    connection.add_setting(settings)

    settings.set_property(nm.SETTING_IP_CONFIG_METHOD, ipv4['method'])
    if (ipv4['method'] == nm.SETTING_IP4_CONFIG_METHOD_MANUAL or
        ipv4['method'] == nm.SETTING_IP4_CONFIG_METHOD_SHARED) and \
       ipv4['address']:
        ipv4_address_int = ipv4_string_to_int(ipv4['address'])

        if not ipv4['netmask']:
            ipv4_netmask_int = nm.utils_ip4_get_default_prefix(
                ipv4_address_int)
        else:
            ipv4_netmask_int = nm.utils_ip4_netmask_to_prefix(
                ipv4_string_to_int(ipv4['netmask']))

        address = nm.IPAddress.new(socket.AF_INET, ipv4['address'],
                                   ipv4_netmask_int)
        settings.add_address(address)

        if not ipv4['gateway']:
            settings.set_property(nm.SETTING_IP_CONFIG_GATEWAY, '0.0.0.0')
        else:
            settings.set_property(nm.SETTING_IP_CONFIG_GATEWAY,
                                  ipv4['gateway'])
    else:
        if ipv4['dns'] or ipv4['second_dns']:
            settings.set_property(nm.SETTING_IP_CONFIG_IGNORE_AUTO_DNS, True)

    if ipv4['dns']:
        settings.add_dns(ipv4['dns'])

    if ipv4['second_dns']:
        settings.add_dns(ipv4['second_dns'])


def _update_ipv6_settings(connection, ipv6):
    """Edit IPv6 settings for network manager connections."""
    settings = nm.SettingIP6Config.new()
    connection.add_setting(settings)

    settings.set_property(nm.SETTING_IP_CONFIG_METHOD, ipv6['method'])
    if ipv6['method'] == nm.SETTING_IP6_CONFIG_METHOD_MANUAL and \
       ipv6['address'] and ipv6['prefix']:
        address = nm.IPAddress.new(socket.AF_INET6, ipv6['address'],
                                   int(ipv6['prefix']))
        settings.add_address(address)

        if not ipv6['gateway']:
            settings.set_property(nm.SETTING_IP_CONFIG_GATEWAY, '::')
        else:
            settings.set_property(nm.SETTING_IP_CONFIG_GATEWAY,
                                  ipv6['gateway'])
    else:
        if ipv6['dns'] or ipv6['second_dns']:
            settings.set_property(nm.SETTING_IP_CONFIG_IGNORE_AUTO_DNS, True)

    if ipv6['dns']:
        settings.add_dns(ipv6['dns'])

    if ipv6['second_dns']:
        settings.add_dns(ipv6['second_dns'])


def _update_pppoe_settings(connection, pppoe):
    """Create/edit PPPoE settings for network manager connections."""
    # PPPoE
    settings = connection.get_setting_pppoe()
    if not settings:
        settings = nm.SettingPppoe.new()
        connection.add_setting(settings)

    settings.set_property(nm.SETTING_PPPOE_USERNAME, pppoe['username'])
    settings.set_property(nm.SETTING_PPPOE_PASSWORD, pppoe['password'])

    settings = connection.get_setting_ppp()
    if not settings:
        settings = nm.SettingPpp.new()
        connection.add_setting(settings)

    # TODO: make this configurable? Some PPP peers don't respond to
    # echo requests according to NetworkManager documentation.
    settings.set_property(nm.SETTING_PPP_LCP_ECHO_FAILURE, 5)
    settings.set_property(nm.SETTING_PPP_LCP_ECHO_INTERVAL, 30)

    return connection


def _update_wireless_settings(connection, wireless):
    """Create/edit wifi settings for network manager connections."""
    key_mgmt = 'wpa-psk'

    # Wireless
    settings = connection.get_setting_wireless()
    if not settings:
        settings = nm.SettingWireless.new()
        connection.add_setting(settings)

    ssid_gbytes = glib.Bytes.new(wireless['ssid'].encode())
    settings.set_property(nm.SETTING_WIRELESS_SSID, ssid_gbytes)
    settings.set_property(nm.SETTING_WIRELESS_MODE, wireless['mode'])
    band = wireless['band'] if wireless['band'] != 'auto' else None
    settings.set_property(nm.SETTING_WIRELESS_BAND, band)
    channel = wireless['channel']
    if wireless['band'] == 'auto' or not wireless['channel']:
        channel = 0

    settings.set_property(nm.SETTING_WIRELESS_CHANNEL, channel)
    settings.set_property(nm.SETTING_WIRELESS_BSSID, wireless['bssid'] or None)

    # Wireless Security
    if wireless['auth_mode'] == 'wpa' and wireless['passphrase']:
        settings = connection.get_setting_wireless_security()
        if not settings:
            settings = nm.SettingWirelessSecurity.new()
            connection.add_setting(settings)

        settings.set_property(nm.SETTING_WIRELESS_SECURITY_KEY_MGMT, key_mgmt)
        settings.set_property(nm.SETTING_WIRELESS_SECURITY_PSK,
                              wireless['passphrase'])
    else:
        connection.remove_setting(nm.SettingWirelessSecurity)

    return connection


def _update_wireguard_settings(connection, wireguard):
    """Create/edit WireGuard settings for network manager connections."""
    settings = connection.get_setting_by_name('wireguard')
    if not settings:
        settings = nm.SettingWireGuard.new()
        connection.add_setting(settings)

    settings.set_property(nm.SETTING_WIREGUARD_PRIVATE_KEY,
                          wireguard['private_key'])
    if 'listen_port' in wireguard:
        settings.set_property(nm.SETTING_WIREGUARD_LISTEN_PORT,
                              wireguard['listen_port'])

    if 'peer_public_key' in wireguard:
        peer = nm.WireGuardPeer.new()
        peer.set_public_key(wireguard['peer_public_key'], False)

        if 'peer_endpoint' in wireguard:
            peer.set_endpoint(wireguard['peer_endpoint'], False)

        if wireguard['preshared_key']:
            # Flag NONE means that NM should store and retain the secret.
            # Default seems to be NOT_REQUIRED in this case.
            peer.set_preshared_key_flags(nm.SettingSecretFlags.NONE)
            peer.set_preshared_key(wireguard['preshared_key'], False)

        for allowed_ip in wireguard['allowed_ips']:
            peer.append_allowed_ip(allowed_ip, False)

        settings.clear_peers()
        settings.append_peer(peer)


def _update_settings(connection, connection_uuid, settings):
    """Create/edit wifi settings for network manager connections."""
    connection = _update_common_settings(connection, connection_uuid,
                                         settings['common'])
    if 'ipv4' in settings and settings['ipv4']:
        _update_ipv4_settings(connection, settings['ipv4'])

    if 'ipv6' in settings and settings['ipv6']:
        _update_ipv6_settings(connection, settings['ipv6'])

    if 'pppoe' in settings and settings['pppoe']:
        _update_pppoe_settings(connection, settings['pppoe'])

    if 'wireless' in settings and settings['wireless']:
        _update_wireless_settings(connection, settings['wireless'])

    if 'wireguard' in settings and settings['wireguard']:
        _update_wireguard_settings(connection, settings['wireguard'])

    return connection


def add_connection(settings):
    """Add an connection in network manager.

    Return the UUID for the connection.
    """
    connection_uuid = str(uuid.uuid4())
    connection = _update_settings(None, connection_uuid, settings)
    client = get_nm_client()
    client.add_connection_async(connection, True, None, _callback, None)
    return connection_uuid


def edit_connection(connection, settings):
    """Edit an existing connection in network manager."""
    _update_settings(connection, connection.get_uuid(), settings)
    connection.commit_changes(True)


def activate_connection(connection_uuid):
    """Find and activate a network connection."""
    connection = get_connection(connection_uuid)
    interface = connection.get_interface_name()
    client = get_nm_client()
    for device in client.get_all_devices():
        if device.get_iface() == interface:
            client.activate_connection_async(connection, device, '/', None,
                                             _callback, None)
            break
    else:
        raise DeviceNotFound(connection)

    return connection


def deactivate_connection(connection_uuid):
    """Find and de-activate a network connection."""
    active_connection = get_active_connection(connection_uuid)
    get_nm_client().deactivate_connection(active_connection)
    return active_connection


def reactivate_connection(connection_uuid):
    """Find and re-activate a network connection to reflect new changes.

    If connection was not active to begin with, do nothing.

    """
    try:
        deactivate_connection(connection_uuid)
    except ConnectionNotFound:
        return  # Connection was not active to start with

    # deactivate_connection() is a synchronous call. However, it returns before
    # the connection is fully deactivated. When re-activating such connections,
    # sometimes, we get a "Authorization request cancelled" error. So, wait
    # until the connection is fully deactivated. XXX: Perform proper
    # asynchronous waiting instead of polling. Also find a way to avoid the
    # problem altogether.
    for index in range(10):  # pylint: disable=unused-variable
        try:
            get_active_connection(connection_uuid)
            time.sleep(0.1)
        except ConnectionNotFound:
            break

    activate_connection(connection_uuid)


def delete_connection(connection_uuid):
    """Delete an exiting connection from network manager.

    Raise ConnectionNotFound if connection does not exist.
    """
    connection = get_connection(connection_uuid)
    name = connection.get_id()
    connection.delete()
    return name


def _get_access_point_as_dict(access_point, active_ap_path):
    """Return Wi-Fi access point information as a dictionary."""
    ssid = access_point.get_ssid()
    if not ssid:  # Hidden network
        return None

    try:
        ssid_string = ssid.get_data().decode(encoding='utf-8')
    except UnicodeError:
        # XXX: Can't deal with binary SSIDs. Don't show SSIDs that are
        # binary only.
        return None

    is_active = (active_ap_path == access_point.get_path())
    return {
        'ssid': ssid,
        'ssid_string': ssid_string,
        'strength': access_point.get_strength(),
        'is_active': is_active
    }


def wifi_scan():
    """Scan for available access points across all Wi-Fi devices."""
    device_access_points = []
    for device in get_nm_client().get_devices():
        if device.get_device_type() != nm.DeviceType.WIFI:
            continue

        if device.get_client() is None:
            # The device got deleted (see NM:show-wifi-networks.py)
            return False

        # Active access point
        active_ap_path = device.get_active_access_point()

        # Last scan time
        last_scan = device.get_last_scan()
        last_scan_time = None
        scan_requested = False
        if last_scan == -1:
            last_scan = None
        else:
            boot_time = time.clock_gettime(time.CLOCK_BOOTTIME)
            last_scan = boot_time - (last_scan / 1000)
            last_scan_time = datetime.datetime.now() - datetime.timedelta(
                seconds=last_scan)

        # Request a scan if the last scan was more than 20 seconds ago
        if (not last_scan) or last_scan > 20:
            device.request_scan()
            scan_requested = True

        # Access points
        access_points = []
        for access_point in device.get_access_points():
            ap_dict = _get_access_point_as_dict(access_point, active_ap_path)
            if ap_dict:
                access_points.append(ap_dict)

        access_points = sorted(access_points,
                               key=lambda point: -point['strength'])

        device_access_points.append({
            'interface_name': device.get_iface(),
            'access_points': access_points,
            'last_scan': last_scan,
            'last_scan_time': last_scan_time,
            'scan_requested': scan_requested
        })

    return sorted(device_access_points,
                  key=lambda device: device['interface_name'])


def refeed_dns():
    """Re-feed DNS servers to systemd-resolved."""
    get_nm_client().reload(nm.ManagerReloadFlags.DNS_RC)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""App component for other apps to manage their STUN/TURN server configuration.
"""

import base64
import hashlib
import hmac
import json
import re
from dataclasses import dataclass, field
from time import time
from typing import ClassVar, Iterable

from plinth import app

TURN_REST_TTL = 24 * 3600

TURN_REST_USER = 'fbxturnuser'

TURN_URI_REGEX = \
    r'(stun:(.*):([0-9]{4}))|(turn:(.*):([0-9]{4})\?transport=(tcp|udp))'


@dataclass
class TurnConfiguration:
    """Data class to hold TURN server configuration.

    domain is the string representing the domain name with which Coturn has
    been configured. This is necessary to associate the correct TLS certificate
    with Coturn communication. STUN/TURN URIs are generated using this.

    shared_secret is a string that must be used by a server to be accepted by
    Coturn server. This is the value set in Coturn configuration file.

    uris are a list of strings that represent the full set of STUN/TURN URIs
    that must be used by a STUN/TURN client after advice from the server.

    """
    domain: str | None = None
    uris: list[str] = field(default_factory=list)
    shared_secret: str | None = None

    def __post_init__(self):
        """Generate URIs after object initialization if necessary."""
        if self.domain and not self.uris:
            self.uris = [
                f'stun:{self.domain}:3478',
                f'turn:{self.domain}:3478?transport=tcp',
                f'turn:{self.domain}:3478?transport=udp'
            ]

    def to_json(self) -> str:
        """Return a JSON representation of the configuration."""
        return json.dumps({
            'domain': self.domain,
            'uris': self.uris,
            'shared_secret': self.shared_secret
        })

    @staticmethod
    def validate_turn_uris(turn_uris: list[str]) -> bool:
        """Return whether the given TURN URI is valid."""
        pattern = re.compile(TURN_URI_REGEX)
        return all(map(pattern.fullmatch, turn_uris))


@dataclass
class UserTurnConfiguration(TurnConfiguration):
    """Data class to hold per-user TURN server configuration.

    username is a string to identify a specific user and is related to
    the credential.

    credential is a string generated for this user that must be used
    by a server to be accepted by Coturn server. It is generated for
    each user separately upon request and will expire after a set
    time.

    """
    username: str | None = None
    credential: str | None = None

    def to_json(self) -> str:
        """Return a JSON representation of the configuration."""
        return json.dumps({
            'domain': self.domain,
            'uris': self.uris,
            'username': self.username,
            'credential': self.credential
        })


class TurnConsumer(app.FollowerComponent):
    """Component to manage coturn configuration.

    In order to provide audio/video calling functionality, communication
    servers very often use an external server such as Coturn for implementing
    the STUN/TURN protocol. To use Coturn, the server needs to be configured
    with a set of URIs provided by Coturn along with a shared secret.

    This component when added to an app allows the app to retrieve the current
    Coturn configuration and respond to any future configuration changes.

    """

    _all: ClassVar[dict[str, 'TurnConsumer']] = {}

    def __init__(self, component_id):
        """Initialize the component.

        component_id should be a unique ID across all components of an app and
        across all components.

        """
        super().__init__(component_id)
        self._all[component_id] = self

    @classmethod
    def list(cls) -> Iterable['TurnConsumer']:
        """Return a list of all Coturn components."""
        return cls._all.values()

    def on_config_change(self, config: TurnConfiguration):
        """Add or update STUN/TURN configuration.

        Override this method and change app's configuration.

        """

    def get_configuration(self) -> TurnConfiguration:
        """Return current coturn configuration."""
        from plinth.modules import coturn
        return coturn.get_config()


class TurnTimeLimitedConsumer(TurnConsumer):
    """Component to manage coturn configuration with time-limited
    credential.

    This component will generate a new credential upon each request,
    which will expire after 1 day. The shared secret is used to
    generate the credential, but is not provided in the configuration.

    """

    def get_configuration(self) -> UserTurnConfiguration:
        """Return user coturn configuration."""
        from plinth.modules import coturn
        static_config = coturn.get_config()
        timestamp = int(time()) + TURN_REST_TTL
        username = str(timestamp) + ':' + TURN_REST_USER
        credential = None
        if static_config.shared_secret:
            digest = hmac.new(bytes(static_config.shared_secret, 'utf-8'),
                              bytes(username, 'utf-8'), hashlib.sha1).digest()
            credential = base64.b64encode(digest).decode()

        return UserTurnConfiguration(static_config.domain, static_config.uris,
                                     None, username, credential)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Expose API over D-Bus to receive renewal notifications.
"""

import logging

from plinth.utils import import_from_gi

gio = import_from_gi('Gio', '2.0')

logger = logging.getLogger(__name__)


class LetsEncrypt:
    """D-Bus service to listen for Let's Encrypt renewal notifications."""

    introspection_xml = '''
<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE node PUBLIC "-//freedesktop//DTD D-BUS Object Introspection 1.0//EN"
  "http://www.freedesktop.org/standards/dbus/1.0/introspect.dtd">
<node name="/org/freedombox/Service/LetsEncrypt">
  <interface name="org.freedombox.Service.LetsEncrypt">
    <method name="DeployHook">
      <arg name="renewed_domains" type="s" direction="in"/>
      <arg name="renewed_lineage" type="s" direction="in"/>
    </method>
  </interface>
</node>
'''

    def register(self, connection):
        """Register the object in D-Bus connection."""
        introspection_data = gio.DBusNodeInfo.new_for_xml(
            self.introspection_xml)
        interface_info = gio.DBusNodeInfo.lookup_interface(
            introspection_data, 'org.freedombox.Service.LetsEncrypt')
        connection.register_object('/org/freedombox/Service/LetsEncrypt',
                                   interface_info, self.on_method_call, None,
                                   None)

    def on_method_call(self, _connection, _sender, _object_path,
                       _interface_name, method_name, parameters, invocation):
        """Handle method being called.

        No need to check all the incoming parameters as D-Bus will validate all
        the incoming parameters using introspection data.

        """
        if method_name == 'DeployHook':
            self._on_deploy_hook(parameters[0], parameters[1])
            invocation.return_value()

    @staticmethod
    def _on_deploy_hook(renewed_domains, renewed_lineage):
        """Run renewal hooks for each app.

        Don't run this a new thread here because
        component.on_certificate_event() creates one anyway.

        """
        logger.info('LE certificate renewed (deployed): %s, %s.',
                    renewed_domains, renewed_lineage)

        from plinth.modules.letsencrypt import components
        components.on_certificate_event('renewed', renewed_domains,
                                        renewed_lineage)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""FreedomBox app to configure SOGo."""

from django.urls import reverse_lazy
from django.utils.translation import gettext_lazy as _

from plinth import app as app_module
from plinth import cfg, frontpage, menu
from plinth.config import DropinConfigs
from plinth.daemon import Daemon, SharedDaemon
from plinth.modules.apache.components import Webserver
from plinth.modules.backups.components import BackupRestore
from plinth.modules.firewall.components import Firewall
from plinth.package import Packages
from plinth.privileged import service as service_privileged
from plinth.utils import format_lazy

from . import manifest, privileged

_description = [
    _('SOGo is a groupware server that provides a rich web interface for '
      'email, calendar, tasks, and contacts. Calendar, tasks, and contacts '
      'can also be accessed with various mobile and desktop applications '
      'using the CalDAV and CardDAV standards.'),
    format_lazy(
        _('Webmail works with the <a href="{email_url}">Postfix/Dovecot</a> '
          'email server app to retrieve, manage, and send email.'),
        email_url=reverse_lazy('email:index')),
    format_lazy(
        _('All users on {box_name} can login into and use SOGo. Mails '
          'delivered to their mailboxes by the email server app can be read '
          'and new mail can be sent out.'), box_name=_(cfg.box_name)),
]


class SOGoApp(app_module.App):
    """FreedomBox app for SOGo."""

    app_id = 'sogo'

    _version = 3

    def __init__(self) -> None:
        """Create components for the app."""
        super().__init__()

        info = app_module.Info(app_id=self.app_id, version=self._version,
                               depends=['email'], name=_('SOGo'),
                               icon_filename='sogo', description=_description,
                               manual_page='SOGo', clients=manifest.clients,
                               tags=manifest.tags)
        self.add(info)

        menu_item = menu.Menu('menu-sogo', info.name, info.icon_filename,
                              info.tags, 'sogo:index', parent_url_name='apps')
        self.add(menu_item)

        shortcut = frontpage.Shortcut('shortcut-sogo', info.name,
                                      icon=info.icon_filename, url='/SOGo/',
                                      clients=info.clients, tags=info.tags)
        self.add(shortcut)

        # dpkg-dev is needed because plget command, part of
        # gnustep-base-runtime package, uses dpkg-architecture command but
        # there is no dependency on dpkg-dev.
        packages = Packages('packages-sogo',
                            ['sogo', 'postgresql', 'memcached', 'dpkg-dev'])
        self.add(packages)

        dropin_configs = DropinConfigs('dropin-configs-sogo', [
            '/etc/apache2/conf-available/sogo-freedombox.conf',
        ])
        self.add(dropin_configs)

        firewall = Firewall('firewall-sogo', info.name,
                            ports=['http', 'https'], is_external=True)
        self.add(firewall)

        webserver = Webserver('webserver-sogo', 'sogo-freedombox',
                              urls=['https://{host}/SOGo/'])
        self.add(webserver)

        daemon1 = SharedDaemon('shared-daemon-sogo-memcached', 'memcached',
                               listen_ports=[(11211, 'tcp4')])
        self.add(daemon1)

        daemon2 = SharedDaemon('shared-daemon-sogo-postgresql', 'postgresql')
        self.add(daemon2)

        daemon3 = Daemon('daemon-sogo', 'sogo', listen_ports=[(20000, 'tcp4')])
        self.add(daemon3)

        backup_restore = SOGoBackupRestore('backup-restore-sogo',
                                           **manifest.backup)
        self.add(backup_restore)

    def setup(self, old_version):
        """Install and configure the app."""
        super().setup(old_version)
        privileged.setup()
        service_privileged.try_restart('sogo')
        service_privileged.try_restart('memcached')

        if not old_version:
            self.enable()

    def uninstall(self):
        """De-configure and uninstall the app."""
        super().uninstall()
        privileged.uninstall()


class SOGoBackupRestore(BackupRestore):
    """Component to backup/restore SOGo."""

    def backup_pre(self, packet):
        """Save database contents."""
        super().backup_pre(packet)
        privileged.dump_database()

    def restore_post(self, packet):
        """Restore database contents."""
        super().restore_post(packet)
        privileged.restore_database()



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Forms for backups module.
"""

import logging
import os
import re
import subprocess

from django import forms
from django.core.exceptions import ValidationError
from django.core.validators import (FileExtensionValidator,
                                    validate_ipv46_address)
from django.utils.translation import gettext
from django.utils.translation import gettext_lazy as _

from plinth import cfg
from plinth.modules.storage import get_mounts
from plinth.utils import format_lazy

from . import api, split_path
from .repository import get_repositories

logger = logging.getLogger(__name__)


def _get_app_choices(components):
    """Return a list of check box multiple choices from list of components."""
    choices = []
    for component in components:
        name = str(component.app.info.name)
        if not component.has_data:
            name = gettext('{app} (No data to backup)').format(
                app=component.app.info.name)

        choices.append((component.app_id, name))

    return sorted(choices, key=lambda choice: choice[1].lower())


def _get_repository_choices():
    """Return the list of available repositories."""
    choices = [(repository.uuid, repository.name)
               for repository in get_repositories() if repository.is_usable()]

    return choices


class ScheduleForm(forms.Form):
    """Form to edit backups schedule."""

    enabled = forms.BooleanField(
        label=_('Enable scheduled backups'), required=False,
        help_text=_('If enabled, a backup is taken every day, every week and '
                    'every month. Older backups are removed.'))

    daily_to_keep = forms.IntegerField(
        label=_('Number of daily backups to keep'), required=True, min_value=0,
        help_text=_('This many latest backups are kept and the rest are '
                    'removed. A value of "0" disables backups of this type. '
                    'Triggered at specified hour every day.'))

    weekly_to_keep = forms.IntegerField(
        label=_('Number of weekly backups to keep'), required=True,
        min_value=0,
        help_text=_('This many latest backups are kept and the rest are '
                    'removed. A value of "0" disables backups of this type. '
                    'Triggered at specified hour every Sunday.'))

    monthly_to_keep = forms.IntegerField(
        label=_('Number of monthly backups to keep'), required=True,
        min_value=0,
        help_text=_('This many latest backups are kept and the rest are '
                    'removed. A value of "0" disables backups of this type. '
                    'Triggered at specified hour first day of every month.'))

    run_at_hour = forms.IntegerField(
        label=_('Hour of the day to trigger backup operation'), required=True,
        min_value=0, max_value=23, help_text=_(
            'In 24 hour format. Services may become temporarily unavailable '
            'while running backup operation at this time of the day.'))

    selected_apps = forms.MultipleChoiceField(
        label=_('Included apps'), help_text=_('Apps to include in the backup'),
        widget=forms.CheckboxSelectMultiple(attrs={'class': 'has-select-all'}))

    def __init__(self, *args, **kwargs):
        """Initialize the form with selectable apps."""
        super().__init__(*args, **kwargs)
        components = api.get_all_components_for_backup()
        choices = _get_app_choices(components)
        self.fields['selected_apps'].choices = choices
        self.fields['selected_apps'].initial = [
            choice[0] for choice in choices
            if choice[0] not in self.initial.get('unselected_apps', [])
        ]


class CreateArchiveForm(forms.Form):
    repository = forms.ChoiceField(label=_('Repository'))
    name = forms.RegexField(
        label=_('Name'),
        help_text=_('(Optional) Set a name for this backup archive'),
        regex=r'^[^{}/]*$', required=False, strip=True)
    selected_apps = forms.MultipleChoiceField(
        label=_('Included apps'), help_text=_('Apps to include in the backup'),
        widget=forms.CheckboxSelectMultiple(attrs={'class': 'has-select-all'}))

    def __init__(self, *args, **kwargs):
        """Initialize the form with selectable apps."""
        super().__init__(*args, **kwargs)
        components = api.get_all_components_for_backup()
        choices = _get_app_choices(components)
        self.fields['selected_apps'].choices = choices
        if not self.initial or 'selected_apps' not in self.initial:
            self.fields['selected_apps'].initial = [
                choice[0] for choice in choices
            ]
        self.fields['repository'].choices = _get_repository_choices()


class RestoreForm(forms.Form):
    selected_apps = forms.MultipleChoiceField(
        label=_('Select the apps you want to restore'),
        widget=forms.CheckboxSelectMultiple(attrs={'class': 'has-select-all'}))

    def __init__(self, *args, **kwargs):
        """Initialize the form with selectable apps."""
        components = kwargs.pop('components')
        super().__init__(*args, **kwargs)
        choices = _get_app_choices(components)
        self.fields['selected_apps'].choices = choices
        self.fields['selected_apps'].initial = [
            choice[0] for choice in choices
        ]


class UploadForm(forms.Form):
    file = forms.FileField(
        label=_('Upload File'), required=True, validators=[
            FileExtensionValidator(
                ['gz'], _('Backup files have to be in .tar.gz format'))
        ], help_text=format_lazy(
            _('Select the backup file to upload from the local computer. This '
              'must be a file previously downloaded from the result of a '
              'successful backup on a {box_name}. It must have a .tar.gz '
              'extension.'), box_name=_(cfg.box_name)))


def repository_validator(path):
    """Validate an SSH repository path."""
    if not ('@' in path and ':' in path):
        raise ValidationError(_('Repository path format incorrect.'))

    username, hostname, dir_path = split_path(path)
    hostname = hostname.split('%')[0]

    # Validate username using Unix username regex
    if not re.match(r'[a-z0-9_][a-z0-9_-]*$', username):
        raise ValidationError(_(f'Invalid username: {username}'))

    # The hostname should either be a valid IP address or hostname
    # Follows RFC1123 (hostnames can start with digits) instead of RFC952
    hostname_re = (r'^(([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-]*[a-zA-Z0-9])\.)*'
                   r'([A-Za-z0-9]|[A-Za-z0-9][A-Za-z0-9\-]*[A-Za-z0-9])$')
    try:
        validate_ipv46_address(hostname)
    except ValidationError:
        if not re.match(hostname_re, hostname):
            raise ValidationError(_(f'Invalid hostname: {hostname}'))

    # Validate directory path
    if not re.match(r'[^\0]*', dir_path):
        raise ValidationError(_(f'Invalid directory path: {dir_path}'))


class EncryptedBackupsMixin(forms.Form):
    """Form to add a new backup repository."""
    encryption = forms.ChoiceField(
        label=_('Encryption'), help_text=format_lazy(
            _('"Key in Repository" means that a '
              'password-protected key is stored with the backup.')),
        choices=[('repokey', _('Key in Repository')), ('none', _('None'))])
    encryption_passphrase = forms.CharField(
        label=_('Passphrase'),
        help_text=_('Only needed when using encryption.'),
        widget=forms.PasswordInput(), required=False)
    confirm_encryption_passphrase = forms.CharField(
        label=_('Confirm Passphrase'), help_text=_('Repeat the passphrase.'),
        widget=forms.PasswordInput(), required=False)

    def clean(self):
        super().clean()
        passphrase = self.cleaned_data.get('encryption_passphrase')
        confirm_passphrase = self.cleaned_data.get(
            'confirm_encryption_passphrase')

        if passphrase != confirm_passphrase:
            raise forms.ValidationError(
                _('The entered encryption passphrases do not match'))

        if self.cleaned_data.get('encryption') != 'none' and not passphrase:
            raise forms.ValidationError(
                _('Passphrase is needed for encryption.'))

        return self.cleaned_data


encryption_fields = [
    'encryption', 'encryption_passphrase', 'confirm_encryption_passphrase'
]


def get_disk_choices():
    """Returns a list of all available partitions except the root partition."""
    repositories = get_repositories()
    existing_paths = [
        repository.path for repository in repositories
        if repository.storage_type == 'disk'
    ]
    choices = []
    for device in get_mounts():
        if device['mount_point'] == '/':
            continue

        path = os.path.join(device['mount_point'], 'FreedomBoxBackups')
        if path in existing_paths:
            continue

        name = device['label'] if device['label'] else device['mount_point']
        choices.append((device['mount_point'], name))

    return choices


class AddRepositoryForm(EncryptedBackupsMixin, forms.Form):
    """Form to create a new backups repository on a disk."""
    disk = forms.ChoiceField(
        label=_('Select Disk or Partition'), help_text=format_lazy(
            _('Backups will be stored in the directory FreedomBoxBackups')),
        choices=get_disk_choices)

    field_order = ['disk'] + encryption_fields


class AddRemoteRepositoryForm(EncryptedBackupsMixin, forms.Form):
    """Form to add new SSH remote repository."""
    repository = forms.CharField(
        label=_('SSH Repository Path'), strip=True,
        help_text=_('Path of a new or existing repository. Example: '
                    '<i>user@host:~/path/to/repo/</i>'),
        validators=[repository_validator])
    ssh_auth_type = forms.ChoiceField(
        label=_('SSH Authentication Type'),
        help_text=_('Choose how to authenticate to the remote SSH server.'),
        widget=forms.RadioSelect(),
        choices=[('key_auth', _('Key-based Authentication')),
                 ('password_auth', _('Password-based Authentication'))])
    ssh_password = forms.CharField(
        label=_('SSH server password'), widget=forms.PasswordInput(),
        strip=True, help_text=_('Required for password-based authentication.'),
        required=False)

    field_order = ['repository', 'ssh_auth_type', 'ssh_password'
                   ] + encryption_fields

    def clean(self):
        """Perform additional checks on form data."""
        super().clean()
        ssh_password = self.cleaned_data.get('ssh_password')
        if self.cleaned_data.get(
                'ssh_auth_type') == 'password_auth' and not ssh_password:
            raise forms.ValidationError(
                _('SSH password is needed for password-based authentication.'))

        return self.cleaned_data

    def clean_repository(self):
        """Validate repository form field."""
        path = self.cleaned_data.get('repository')
        # Avoid creation of duplicate ssh remotes
        self._check_if_duplicate_remote(path)
        return path

    @staticmethod
    def _check_if_duplicate_remote(path):
        """Raise validation error if given path is a stored remote."""
        for repository in get_repositories():
            if repository.path == path:
                raise forms.ValidationError(
                    _('Remote backup repository already exists.'))


class VerifySshHostkeyForm(forms.Form):
    """Form to verify the SSH public key for a host."""
    ssh_public_key = forms.ChoiceField(
        label=_('Select verified SSH public key'), widget=forms.RadioSelect)

    def __init__(self, *args, **kwargs):
        """Initialize the form with selectable apps."""
        hostname = kwargs.pop('hostname')
        super().__init__(*args, **kwargs)
        (self.fields['ssh_public_key'].choices,
         self.keyscan_error) = self._get_all_public_keys(hostname)

    @staticmethod
    def _get_all_public_keys(hostname):
        """Use ssh-keyscan to get all the SSH public keys of a host."""
        # Fetch public keys of ssh remote
        keyscan = subprocess.run(['ssh-keyscan', hostname],
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE, check=False)
        key_lines = keyscan.stdout.decode().splitlines()
        keys = [line for line in key_lines if not line.startswith('#')]
        error_message = keyscan.stderr.decode() if keyscan.returncode else None
        # Generate user-friendly fingerprints of public keys
        keygen = subprocess.run(['ssh-keygen', '-l', '-f', '-'],
                                input=keyscan.stdout, stdout=subprocess.PIPE,
                                check=False)
        fingerprints = keygen.stdout.decode().splitlines()

        return zip(keys, fingerprints), error_message



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Views for the dynamicsdns module."""

import datetime

from django.contrib import messages
from django.shortcuts import redirect
from django.urls import reverse_lazy
from django.utils.translation import gettext_lazy as _
from django.views.generic import TemplateView
from django.views.generic.edit import FormView

from plinth import views
from plinth.modules import dynamicdns

from .forms import DomainForm


class DynamicDNSAppView(views.AppView):
    """View to show app status."""

    app_id = 'dynamicdns'
    template_name = 'dynamicdns.html'

    _error_messages = {
        'timeout': _('Connection timed out'),
        'gaierror': _('Could not find server'),
        'TimeoutError': _('Connection timed out'),
        'ConnectionRefusedError': _('Server refused connection'),
        'ValueError': _('Already up-to-date')
    }

    def get_context_data(self, **kwargs):
        """Return the context data for rendering the template view."""
        context = super().get_context_data(**kwargs)
        config = dynamicdns.get_config()
        context['domains'] = config['domains']

        status = dynamicdns.get_status()
        domains_status = {}
        for domain_name, domain in status['domains'].items():
            if domain_name not in config['domains']:
                continue

            # Create naive datetime object in local timezone
            domain['timestamp'] = datetime.datetime.fromtimestamp(
                domain['timestamp']) if domain['timestamp'] else None
            domains_status[domain_name] = domain
            if domain['error_code'] in self._error_messages:
                domain['error_message'] = self._error_messages[
                    domain['error_code']]

        context['domains_status'] = domains_status
        return context


class DomainView(FormView):
    """View to add/edit a dynamic DNS domain."""

    template_name = 'dynamicdns-domain.html'
    form_class = DomainForm
    prefix = 'domain'
    success_url = reverse_lazy('dynamicdns:index')

    def get_context_data(self, **kwargs):
        """Return the context data for rendering the template view."""
        context = super().get_context_data(**kwargs)
        domain_name = self.kwargs.get('domain')
        if not domain_name:
            context['title'] = _('Add Dynamic Domain')
        else:
            context['title'] = _('Edit Dynamic Domain')

        return context

    def get_initial(self):
        """Get the current values for the form."""
        initial = super().get_initial()
        domain_name = self.kwargs.get('domain')
        domains = dynamicdns.get_config()['domains']
        domain = {}
        if domains and domain_name and domain_name in domains:
            domain = domains[domain_name]

        initial.update(domain)
        return domain

    def form_valid(self, form):
        """Apply the changes submitted in the form."""
        old_status = form.initial
        new_status = form.cleaned_data

        if old_status != new_status:
            try:
                _domain_delete(old_status['domain'])
            except KeyError:
                pass

            _domain_add(new_status['domain'], new_status)
            messages.success(self.request, _('Configuration updated'))

        # Perform an immediate update, even when configuration is not changed.
        dynamicdns.update_dns(None)

        return super().form_valid(form)


def _domain_add(domain: str, domain_config: dict):
    """Add a domain to the configuration."""
    config = dynamicdns.get_config()
    config['domains'][domain] = domain_config
    dynamicdns.set_config(config)
    dynamicdns.notify_domain_added(domain)


def _domain_delete(domain: str):
    """Remove a domain from the configuration.

    Raises KeyError if the domain is not found in the configuration.
    """
    config = dynamicdns.get_config()
    del config['domains'][domain]
    dynamicdns.set_config(config)
    if domain:
        dynamicdns.notify_domain_removed(domain)


class DomainDeleteView(TemplateView):
    """Confirm and delete a domain."""
    template_name = 'dynamicdns-domain-delete.html'

    def get_context_data(self, **kwargs):
        """Return additional context data for rendering the template."""
        context = super().get_context_data(**kwargs)
        domain = self.kwargs['domain']
        context['domain'] = domain
        context['title'] = str(
            _('Delete Domain {domain}?')).format(domain=domain)
        return context

    def post(self, request, domain):
        """Delete a domain."""
        try:
            _domain_delete(domain)
            messages.success(request, _('Domain deleted.'))
        except KeyError:
            raise

        return redirect('dynamicdns:index')



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Test module for custom context processors.
"""

from unittest.mock import MagicMock, Mock, patch

from django.urls import resolve

from plinth import context_processors as cp


@patch('plinth.notification.Notification')
def test_common(Notification, load_cfg, rf, test_menu):
    """Verify that the common() function returns the correct values."""
    url = '/apps/testapp/create/'
    request = rf.get(url)
    request.resolver_match = resolve(url)
    request.user = Mock()
    request.user.groups.filter().exists = Mock(return_value=True)
    request.session = MagicMock()
    response = cp.common(request)
    assert response is not None

    config = response['cfg']
    assert config is not None
    assert config.box_name == 'FreedomBox'
    assert response['box_name'] == 'FreedomBox'
    assert len(response['breadcrumbs']) == 4
    assert response['active_section_url'] == '/apps/'
    assert response['user_is_admin']



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Configure Tor service."""

import codecs
import logging
import os
import pathlib
import re
import shutil
import socket
import time
from typing import Any

import augeas

from plinth import action_utils
from plinth.actions import privileged

INSTANCE_NAME = 'plinth'
SERVICE_FILE = '/etc/firewalld/services/tor-{0}.xml'
SERVICE_NAME = f'tor@{INSTANCE_NAME}'
TOR_CONFIG = f'/etc/tor/instances/{INSTANCE_NAME}/torrc'
TOR_CONFIG_AUG = f'/files/{TOR_CONFIG}'
TOR_STATE_FILE = f'/var/lib/tor-instances/{INSTANCE_NAME}/state'
TOR_AUTH_COOKIE = f'/var/run/tor-instances/{INSTANCE_NAME}/control.authcookie'
TOR_APACHE_SITE = '/etc/apache2/conf-available/onion-location-freedombox.conf'

logger = logging.getLogger(__name__)


@privileged
def setup(old_version: int):
    """Setup Tor configuration after installing it."""
    # Disable default Tor service. We will use tor@plinth instance instead.
    action_utils.service_disable('tor@default')
    # Mask the service to prevent re-enabling it by the Tor master service.
    action_utils.service_mask('tor@default')

    if old_version:
        if old_version <= 4:
            _upgrade_orport_value()

        if old_version <= 6:
            _remove_proxy()

        return

    _first_time_setup()
    _set_onion_header(_get_hidden_service(aug=None))


def _first_time_setup():
    """Setup Tor configuration for the first time setting defaults."""
    logger.info('Performing first time setup for Tor')

    action_utils.run(['tor-instance-create', INSTANCE_NAME], check=True)

    # Remove line starting with +SocksPort, since our augeas lens
    # doesn't handle it correctly.
    with open(TOR_CONFIG, 'r', encoding='utf-8') as torrc:
        torrc_lines = torrc.readlines()
    with open(TOR_CONFIG, 'w', encoding='utf-8') as torrc:
        for line in torrc_lines:
            if not line.startswith('+'):
                torrc.write(line)

    aug = augeas_load()

    aug.set(TOR_CONFIG_AUG + '/ControlPort', '9051')
    _enable_relay(relay=True, bridge=True, aug=aug)
    aug.set(TOR_CONFIG_AUG + '/ExitPolicy[1]', 'reject *:*')
    aug.set(TOR_CONFIG_AUG + '/ExitPolicy[2]', 'reject6 *:*')

    aug.set(TOR_CONFIG_AUG + '/HiddenServiceDir',
            f'/var/lib/tor-instances/{INSTANCE_NAME}/hidden_service')
    aug.set(TOR_CONFIG_AUG + '/HiddenServicePort[1]', '22 127.0.0.1:22')
    aug.set(TOR_CONFIG_AUG + '/HiddenServicePort[2]', '80 127.0.0.1:80')
    aug.set(TOR_CONFIG_AUG + '/HiddenServicePort[3]', '443 127.0.0.1:443')

    aug.save()

    action_utils.service_enable(SERVICE_NAME)
    action_utils.service_restart(SERVICE_NAME)
    _update_ports()

    # wait until hidden service information is available
    tries = 0
    while not _get_hidden_service()['enabled']:
        tries += 1
        if tries >= 12:
            return

        time.sleep(10)


def _upgrade_orport_value():
    """Change ORPort value from auto to 9001.

    When ORPort is set to 'auto', Tor automatically allocates a port for it.
    During it's first run, we able to extract the port number and open the
    firewall port. However, unlike for pluggable transports, Tor does not seem
    to store this port for future reuse in the state file. It hence opens a new
    port every time it is started. This leads to a new port being assigned on
    next Tor startup and leads to relay functionality not being reachable from
    outside.

    According to the documentation, only possible values for ORPort are a fixed
    number or 0 (disable) or auto (above behavior). Choose 9001 as this is
    the commonly used port number for ORPort. The recommended port number of
    443 is not possible in FreedomBox due it is use for other purposes.

    """
    logger.info('Upgrading ORPort value for Tor')
    aug = augeas_load()

    if _is_relay_enabled(aug):
        aug.set(TOR_CONFIG_AUG + '/ORPort[1]', '9001')
        aug.set(TOR_CONFIG_AUG + '/ORPort[2]', '[::]:9001')

    aug.save()

    action_utils.service_try_restart(SERVICE_NAME)

    # Tor may not be running, don't try to read/update all ports
    _update_port('orport', 9001)
    action_utils.service_restart('firewalld')


def _remove_proxy():
    """Remove SocksProxy from configuration.

    This functionality was split off to a separate app, Tor Proxy.
    """
    logger.info('Removing SocksProxy from Tor configuration')
    aug = augeas_load()
    for config in [
            'SocksPort', 'VirtualAddrNetworkIPv4', 'AutomapHostsOnResolve',
            'TransPort', 'DNSPort'
    ]:
        aug.remove(TOR_CONFIG_AUG + '/' + config)

    aug.save()
    action_utils.service_try_restart(SERVICE_NAME)


@privileged
def configure(use_upstream_bridges: bool | None = None,
              upstream_bridges: str | None = None, relay: bool | None = None,
              bridge_relay: bool | None = None,
              hidden_service: bool | None = None):
    """Configure Tor."""
    aug = augeas_load()

    _use_upstream_bridges(use_upstream_bridges, aug=aug)

    if use_upstream_bridges:
        relay = False
        bridge_relay = False

    if upstream_bridges:
        _set_upstream_bridges(upstream_bridges, aug=aug)

    _enable_relay(relay, bridge_relay, aug=aug)

    if hidden_service:
        _enable_hs(aug=aug)
    elif hidden_service is not None:
        _disable_hs(aug=aug)


@privileged
def update_ports():
    """Update firewall ports based on what Tor uses."""
    _update_ports()


@privileged
def restart():
    """Restart Tor."""
    if (action_utils.service_is_enabled(SERVICE_NAME, strict_check=True)
            and action_utils.service_is_running(SERVICE_NAME)):
        action_utils.service_restart(SERVICE_NAME)

        aug = augeas_load()
        if aug.get(TOR_CONFIG_AUG + '/HiddenServiceDir'):
            # wait until hidden service information is available
            tries = 0
            while not _get_hidden_service()['enabled']:
                tries += 1
                if tries >= 12:
                    return
                time.sleep(10)


@privileged
def get_status() -> dict[str, bool | str | dict[str, Any]]:
    """Return dict with Tor status."""
    aug = augeas_load()
    return {
        'use_upstream_bridges': _are_upstream_bridges_enabled(aug),
        'upstream_bridges': _get_upstream_bridges(aug),
        'relay_enabled': _is_relay_enabled(aug),
        'bridge_relay_enabled': _is_bridge_relay_enabled(aug),
        'ports': _get_ports(),
        'hidden_service': _get_hidden_service(aug)
    }


def _are_upstream_bridges_enabled(aug) -> bool:
    """Return whether upstream bridges are being used."""
    use_bridges = aug.get(TOR_CONFIG_AUG + '/UseBridges')
    return use_bridges == '1'


def _get_upstream_bridges(aug) -> str:
    """Return upstream bridges separated by newlines."""
    matches = aug.match(TOR_CONFIG_AUG + '/Bridge')
    bridges = [aug.get(match) for match in matches]
    return '\n'.join(bridges)


def _is_relay_enabled(aug) -> bool:
    """Return whether a relay is enabled."""
    orport = aug.get(TOR_CONFIG_AUG + '/ORPort[1]')
    return bool(orport) and orport != '0'


def _is_bridge_relay_enabled(aug) -> bool:
    """Return whether bridge relay is enabled."""
    bridge = aug.get(TOR_CONFIG_AUG + '/BridgeRelay')
    return bridge == '1'


def _get_ports() -> dict[str, str]:
    """Return dict mapping port names to numbers."""
    ports = {}
    try:
        ports['orport'] = _get_orport()
    except Exception:
        pass

    try:
        with open(TOR_STATE_FILE, 'r', encoding='utf-8') as state_file:
            for line in state_file:
                matches = re.match(
                    r'^\s*TransportProxy\s+(\S*)\s+\S+:(\d+)\s*$', line)
                if matches:
                    ports[matches.group(1)] = matches.group(2)
    except FileNotFoundError:
        pass

    return ports


def _get_orport() -> str:
    """Return the ORPort by querying running instance."""
    cookie_bytes = open(TOR_AUTH_COOKIE, 'rb').read()
    cookie = codecs.encode(cookie_bytes, 'hex').decode()

    commands = '''AUTHENTICATE {cookie}
GETINFO net/listeners/or
QUIT
'''.format(cookie=cookie)

    tor_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    tor_socket.connect(('localhost', 9051))
    tor_socket.send(commands.encode())
    response = tor_socket.recv(1024)
    tor_socket.close()

    line = response.split(b'\r\n')[1].decode()
    matches = re.match(r'.*=".+:(\d+)"', line)
    if not matches:
        raise ValueError('Invalid orport value returned by Tor')

    return matches.group(1)


def _get_hidden_service(aug=None) -> dict[str, Any]:
    """Return a string with configured Tor hidden service information."""
    hs_enabled = False
    hs_status = 'Ok'
    hs_hostname = None
    hs_ports = []

    if not aug:
        aug = augeas_load()

    hs_dir = aug.get(TOR_CONFIG_AUG + '/HiddenServiceDir')
    hs_port_paths = aug.match(TOR_CONFIG_AUG + '/HiddenServicePort')

    for hs_port_path in hs_port_paths:
        port_info = aug.get(hs_port_path).split()
        hs_ports.append({'virtport': port_info[0], 'target': port_info[1]})

    if not hs_dir:
        hs_status = 'Not Configured'
    else:
        try:
            with open(os.path.join(hs_dir, 'hostname'), 'r',
                      encoding='utf-8') as conf_file:
                hs_hostname = conf_file.read().strip()
                hs_enabled = True
        except Exception:
            hs_status = 'Not available (Run Tor at least once)'

    return {
        'enabled': hs_enabled,
        'status': hs_status,
        'hostname': hs_hostname,
        'ports': hs_ports
    }


def _use_upstream_bridges(use_upstream_bridges: bool | None = None, aug=None):
    """Enable use of upstream bridges."""
    if use_upstream_bridges is None:
        return

    if not aug:
        aug = augeas_load()

    if use_upstream_bridges:
        aug.set(TOR_CONFIG_AUG + '/UseBridges', '1')
    else:
        aug.set(TOR_CONFIG_AUG + '/UseBridges', '0')

    aug.save()


def _set_upstream_bridges(upstream_bridges=None, aug=None):
    """Set list of upstream bridges."""
    if upstream_bridges is None:
        return

    if not aug:
        aug = augeas_load()

    aug.remove(TOR_CONFIG_AUG + '/Bridge')
    if upstream_bridges:
        bridges = [bridge.strip() for bridge in upstream_bridges.split('\n')]
        bridges = [bridge for bridge in bridges if bridge]
        for bridge in bridges:
            parts = [part for part in bridge.split() if part]
            bridge = ' '.join(parts)
            aug.set(TOR_CONFIG_AUG + '/Bridge[last() + 1]', bridge.strip())

    aug.set(TOR_CONFIG_AUG + '/ClientTransportPlugin',
            'obfs3,scramblesuit,obfs4 exec /usr/bin/obfs4proxy')

    aug.save()


def _enable_relay(relay: bool | None, bridge: bool | None, aug: augeas.Augeas):
    """Enable Tor bridge relay."""
    if relay is None and bridge is None:
        return

    if not aug:
        aug = augeas_load()

    use_upstream_bridges = _are_upstream_bridges_enabled(aug)

    if relay and not use_upstream_bridges:
        aug.set(TOR_CONFIG_AUG + '/ORPort[1]', '9001')
        aug.set(TOR_CONFIG_AUG + '/ORPort[2]', '[::]:9001')
    elif relay is not None:
        aug.remove(TOR_CONFIG_AUG + '/ORPort')

    if bridge and not use_upstream_bridges:
        aug.set(TOR_CONFIG_AUG + '/BridgeRelay', '1')
        aug.set(TOR_CONFIG_AUG + '/ServerTransportPlugin',
                'obfs3,obfs4 exec /usr/bin/obfs4proxy')
        aug.set(TOR_CONFIG_AUG + '/ExtORPort', 'auto')
    elif bridge is not None:
        aug.remove(TOR_CONFIG_AUG + '/BridgeRelay')
        aug.remove(TOR_CONFIG_AUG + '/ServerTransportPlugin')
        aug.remove(TOR_CONFIG_AUG + '/ExtORPort')

    aug.save()


def _enable_hs(aug=None):
    """Enable Tor hidden service."""
    if not aug:
        aug = augeas_load()

    if _get_hidden_service(aug)['enabled']:
        return

    aug.set(TOR_CONFIG_AUG + '/HiddenServiceDir',
            f'/var/lib/tor-instances/{INSTANCE_NAME}/hidden_service')
    aug.set(TOR_CONFIG_AUG + '/HiddenServicePort[1]', '22 127.0.0.1:22')
    aug.set(TOR_CONFIG_AUG + '/HiddenServicePort[2]', '80 127.0.0.1:80')
    aug.set(TOR_CONFIG_AUG + '/HiddenServicePort[3]', '443 127.0.0.1:443')
    aug.save()
    _set_onion_header(_get_hidden_service(aug))


def _disable_hs(aug=None):
    """Disable Tor hidden service."""
    if not aug:
        aug = augeas_load()

    if not _get_hidden_service(aug)['enabled']:
        return

    aug.remove(TOR_CONFIG_AUG + '/HiddenServiceDir')
    aug.remove(TOR_CONFIG_AUG + '/HiddenServicePort')
    aug.save()
    _set_onion_header(None)


def _update_port(name, number):
    """Update firewall service information for single port."""
    lines = """<?xml version="1.0" encoding="utf-8"?>
<service>
  <short>Tor - {0}</short>
  <port protocol="tcp" port="{1}"/>
</service>
"""
    try:
        with open(SERVICE_FILE.format(name), 'w',
                  encoding='utf-8') as service_file:
            service_file.writelines(lines.format(name, number))
    except FileNotFoundError:
        return


def _update_ports():
    """Update firewall service information."""
    ready = False
    tries = 0

    # port information may not be available immediately after Tor started
    while not ready:
        ports = _get_ports()
        ready = 'orport' in ports and 'obfs3' in ports and 'obfs4' in ports
        if ready:
            break

        tries += 1
        if tries >= 12:
            return

        time.sleep(10)

    for name, number in ports.items():
        _update_port(name, number)

    # XXX: We should ideally do firewalld reload instead.  However,
    # firewalld seems to fail to successfully reload sometimes.
    action_utils.service_restart('firewalld')


def augeas_load():
    """Initialize Augeas."""
    aug = augeas.Augeas(flags=augeas.Augeas.NO_LOAD +
                        augeas.Augeas.NO_MODL_AUTOLOAD)
    aug.set('/augeas/load/Tor/lens', 'Tor.lns')
    aug.set('/augeas/load/Tor/incl[last() + 1]', TOR_CONFIG)
    aug.load()
    return aug


def _set_onion_header(hidden_service):
    """Set Apache configuration for the Onion-Location header."""
    logger.info('Setting Onion-Location header for Apache')
    config_file = pathlib.Path(TOR_APACHE_SITE)
    if hidden_service and hidden_service['enabled']:
        # https://community.torproject.org/onion-services/advanced/onion-location/
        hostname = hidden_service['hostname']
        config_contents = f'''# This file is managed by FreedomBox
<LocationMatch "^(?!/(wordpress|mediawiki))/[^/]+">
    Header set Onion-Location "http://{hostname}%{{REQUEST_URI}}s"
</LocationMatch>
'''
        config_file.write_text(config_contents, encoding='utf-8')
    else:
        config_file.write_text('# This file is managed by FreedomBox\n',
                               encoding='utf-8')

    action_utils.service_reload('apache2')


@privileged
def uninstall():
    """Remove plinth instance."""
    directories = [
        f'/etc/tor/instances/{INSTANCE_NAME}/',
        f'/var/lib/tor-instances/{INSTANCE_NAME}/',
        f'/var/run/tor-instances/{INSTANCE_NAME}/'
    ]
    for directory in directories:
        shutil.rmtree(directory, ignore_errors=True)

    os.unlink(f'/var/run/tor-instances/{INSTANCE_NAME}.defaults')
    action_utils.service_unmask(SERVICE_NAME)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""FreedomBox app for configuring OpenVPN server."""

import logging

from django.http import HttpResponse

from plinth.modules.names.components import DomainName
from plinth.views import AppView

from . import privileged

logger = logging.getLogger(__name__)


class OpenVPNAppView(AppView):
    """Show OpenVPN app main page."""

    app_id = 'openvpn'
    template_name = 'openvpn.html'


def profile(request):
    """Provide the user's profile for download."""
    username = request.user.username
    domain_name = DomainName.list_names()[0]

    profile_string = privileged.get_profile(username, domain_name)
    response = HttpResponse(profile_string,
                            content_type='application/x-openvpn-profile')
    response['Content-Disposition'] = \
        'attachment; filename={username}.ovpn'.format(username=username)

    return response



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Functional, browser based tests for Home Assistant app."""

import pytest

from plinth.tests import functional

pytestmark = [pytest.mark.apps, pytest.mark.homeassistant]


class TestHomeAssistantApp(functional.BaseAppTests):
    """Basic tests for Home Assistant app."""
    app_name = 'homeassistant'
    has_service = True
    has_web = False  # Can't yet check separate sub-domain
    diagnostics_delay = 5

    def install_and_setup(self, session_browser):
        """Set the domain to freedombox.local so that it can tested."""
        super().install_and_setup(session_browser)
        _domain_set(session_browser, 'freedombox.local')


def _domain_set(browser, domain):
    """Set the domain in the domain selection drop down."""
    functional.nav_to_module(browser, 'homeassistant')
    browser.select('domain_name', domain)
    functional.submit(browser, form_class='form-configuration')



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Utilities to help with MariaDB/MySQL databases.

Uses utilities from 'mysql-client' package such as 'mysql' and 'mysqldump'.
"""

import subprocess

from .. import action_utils


def run_query(database_name: str, query: str) -> subprocess.CompletedProcess:
    """Run a database query using 'root' user.

    Does not ensure that the database server is running.
    """
    return action_utils.run(
        ['mysql', '--user=root', '--database', database_name],
        input=query.encode('utf-8'), check=True)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""FreedomBox app to manage backup archives."""

import collections.abc
import contextlib
import json
import logging
import os
import re
import subprocess
from pathlib import Path

from django.utils.text import get_valid_filename
from django.utils.translation import gettext_lazy as _
from django.utils.translation import gettext_noop

from plinth import app as app_module
from plinth import cfg, glib, menu
from plinth.modules.names.components import DomainName
from plinth.package import Packages

from . import api, errors, manifest, privileged

logger = logging.getLogger(__name__)

_description = [
    _('Backups allows creating and managing backup archives.'),
]

# session variable name that stores when a backup file should be deleted
SESSION_PATH_VARIABLE = 'fbx-backups-upload-path'


class BackupsApp(app_module.App):
    """FreedomBox app for backup and restore."""

    app_id = 'backups'

    _version = 4

    can_be_disabled = False

    def __init__(self) -> None:
        """Create components for the app."""
        super().__init__()

        info = app_module.Info(
            app_id=self.app_id, version=self._version, is_essential=True,
            depends=['storage'], name=_('Backups'), icon='fa-files-o',
            description=_description, manual_page='Backups',
            donation_url='https://www.borgbackup.org/support/fund.html',
            tags=manifest.tags)
        self.add(info)

        menu_item = menu.Menu('menu-backups', info.name, info.icon, info.tags,
                              'backups:index', parent_url_name='system:data',
                              order=20)
        self.add(menu_item)

        packages = Packages('packages-backups',
                            ['borgbackup', 'sshfs', 'sshpass'])
        self.add(packages)

    @staticmethod
    def post_init():
        """Perform post initialization operations."""
        # Check every hour to perform scheduled backups
        glib.schedule(3600, backup_by_schedule)

    def setup(self, old_version):
        """Install and configure the app."""
        super().setup(old_version)
        from . import repository
        privileged.setup(repository.RootBorgRepository.PATH)
        self.enable()

        # First time setup or upgrading from older versions.
        if old_version <= 2:
            _show_schedule_setup_notification()


def _backup_handler(packet, encryption_passphrase=None):
    """Perform backup operation on packet."""
    if not os.path.exists(privileged.MANIFESTS_FOLDER):
        os.makedirs(privileged.MANIFESTS_FOLDER)

    manifest_path = os.path.join(privileged.MANIFESTS_FOLDER,
                                 get_valid_filename(packet.path) + '.json')
    manifests = {
        'apps': [{
            'name': component.app.app_id,
            'version': component.app.info.version,
            'backup': component.manifest
        } for component in packet.components]
    }
    with open(manifest_path, 'w', encoding='utf-8') as manifest_file:
        json.dump(manifests, manifest_file)

    paths = packet.directories + packet.files
    paths.append(manifest_path)

    privileged.create_archive(packet.path, paths,
                              comment=packet.archive_comment,
                              encryption_passphrase=encryption_passphrase)


def backup_by_schedule(data):
    """Check if backups need to be taken and run the operation."""
    from . import repository as repository_module
    for repository in repository_module.get_repositories():
        try:
            repository.schedule.run_schedule()
            _show_schedule_error_notification(repository, is_error=False)
        except Exception as exception:
            logger.exception('Error running scheduled backup: %s', exception)
            _show_schedule_error_notification(repository, is_error=True,
                                              exception=exception)


def _restore_exported_archive_handler(packet, encryption_passphrase=None):
    """Perform restore operation on packet."""
    privileged.restore_exported_archive(packet.path, packet.directories,
                                        packet.files)


def restore_archive_handler(packet, encryption_passphrase=None):
    """Perform restore operation on packet."""
    privileged.restore_archive(packet.path, '/', packet.directories,
                               packet.files, encryption_passphrase)


def restore_from_upload(path, app_ids=None):
    """Restore files from an uploaded .tar.gz backup file"""
    api.restore_apps(_restore_exported_archive_handler, app_ids=app_ids,
                     create_subvolume=False, backup_file=path)


def get_known_hosts_path() -> Path:
    """Return the path to the known hosts file."""
    return Path(cfg.data_dir) / '.ssh' / 'known_hosts'


def get_ssh_client_auth_key_paths() -> tuple[Path, Path]:
    """Return the paths to the SSH client public key and private key."""
    key_path = Path(cfg.data_dir) / '.ssh' / 'id_ed25519'
    pubkey_path = key_path.with_suffix('.pub')
    return pubkey_path, key_path


def generate_ssh_client_auth_key():
    """Generate SSH client authentication keypair, if needed."""
    _, key_path = get_ssh_client_auth_key_paths()
    key_path.parent.mkdir(parents=True, exist_ok=True)
    if key_path.exists():
        logger.info('SSH client key %s for FreedomBox service already exists',
                    key_path)
        return

    domain_name = DomainName.list_names()[0]
    logger.info('Generating SSH client key %s for FreedomBox service',
                key_path)
    subprocess.run([
        'ssh-keygen', '-t', 'ed25519', '-N', '', '-C',
        f'freedombox@{domain_name}', '-f',
        str(key_path)
    ], stdout=subprocess.DEVNULL, check=True)


def get_ssh_client_public_key() -> str:
    """Get SSH client public key for FreedomBox service."""
    pubkey_path, _ = get_ssh_client_auth_key_paths()
    with pubkey_path.open('r') as pubkey_file:
        pubkey = pubkey_file.read()

    return pubkey


def copy_ssh_client_public_key(pubkey_path: str, hostname: str, username: str,
                               password: str):
    """Copy the SSH client public key to the remote server.

    Returns whether the copy was successful, and any error message.
    """
    env = os.environ.copy()
    env['SSHPASS'] = str(password)
    with raise_ssh_error():
        try:
            subprocess.run([
                'sshpass', '-e', 'ssh-copy-id', '-i', pubkey_path,
                f'{username}@{hostname}'
            ], stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=True,
                           env=env)
            logger.info("Copied SSH client public key to remote host's "
                        "authorized keys.")
        except subprocess.CalledProcessError as exception:
            logger.warning('Failed to copy SSH client public key: %s',
                           exception.stderr)
            raise


@contextlib.contextmanager
def raise_ssh_error() -> collections.abc.Generator[None]:
    """Convert subprocess error to SshError."""
    try:
        yield
    except subprocess.CalledProcessError as exception:
        raise errors.SshError(exception.returncode, exception.cmd,
                              exception.output, exception.stderr)


def is_ssh_hostkey_verified(hostname):
    """Check whether SSH Hostkey has already been verified.

    hostname: Domain name or IP address of the host

    """
    known_hosts_path = get_known_hosts_path()
    if not known_hosts_path.exists():
        return False

    try:
        subprocess.run(
            ['ssh-keygen', '-F', hostname, '-f',
             str(known_hosts_path)], check=True)
        return True
    except subprocess.CalledProcessError:
        return False


def split_path(path):
    """Splits the given path into username, hostname, directory.

    Network interface information is kept in the hostname if provided.
    e.g. fe80::2078:6c26:498a:1fa5%wlp1s0

    """
    return re.findall(r'^(.*)@([^/]*):(.*)$', path)[0]


def _show_schedule_setup_notification():
    """Show a notification hinting to setup a remote backup schedule."""
    from plinth.notification import Notification
    message = gettext_noop(
        'Enable an automatic backup schedule for data safety. Prefer an '
        'encrypted remote backup location or an extra attached disk.')
    data = {
        'app_name': 'translate:' + gettext_noop('Backups'),
        'app_icon': 'fa-files-o'
    }
    title = gettext_noop('Enable a Backup Schedule')
    actions_ = [{
        'type': 'link',
        'class': 'primary',
        'text': gettext_noop('Go to {app_name}'),
        'url': 'backups:index'
    }, {
        'type': 'dismiss'
    }]
    Notification.update_or_create(id='backups-remote-schedule',
                                  app_id='backups', severity='info',
                                  title=title, message=message,
                                  actions=actions_, data=data, group='admin')


def on_schedule_save(repository):
    """Dismiss notification. Called when repository's schedule is updated."""
    if not repository.schedule.enabled:
        return

    from plinth.notification import Notification
    try:
        note = Notification.get('backups-remote-schedule')
        note.dismiss()
    except KeyError:
        pass


def _show_schedule_error_notification(repository, is_error, exception=None):
    """Show or hide a notification related scheduled backup operation."""
    from plinth.notification import Notification
    id_ = 'backups-schedule-error-' + repository.uuid
    try:
        note = Notification.get(id_)
        error_count = note.data['error_count']
    except KeyError:
        error_count = 0

    message = gettext_noop(
        'A scheduled backup failed. Past {error_count} attempts for backup '
        'did not succeed. The latest error is: {error_message}')
    data = {
        'app_name': 'translate:' + gettext_noop('Backups'),
        'app_icon': 'fa-files-o',
        'error_count': error_count + 1 if is_error else 0,
        'error_message': str(exception)
    }
    title = gettext_noop('Error During Backup')
    actions_ = [{
        'type': 'link',
        'class': 'primary',
        'text': gettext_noop('Go to {app_name}'),
        'url': 'backups:index'
    }, {
        'type': 'dismiss'
    }]
    note = Notification.update_or_create(id=id_, app_id='backups',
                                         severity='error', title=title,
                                         message=message, actions=actions_,
                                         data=data, group='admin')
    note.dismiss(should_dismiss=not is_error)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Application manifest for diagnostics.
"""

from django.utils.translation import gettext_lazy as _

backup: dict = {}

tags = [_('Detect problems'), _('Repair'), _('Daily')]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Views for the Janus app.
"""

import copy

from django.views.generic import TemplateView

from plinth import app as app_module
from plinth.middleware import CONTENT_SECURITY_POLICY


class JanusRoomView(TemplateView):
    """A simple page to host Janus video room."""
    template_name = 'janus_video_room.html'
    headers: dict[str, str] = {}

    def __init__(self, **kwargs):
        """Initialize the view and set CSP."""
        super().__init__(**kwargs)
        csp = copy.copy(CONTENT_SECURITY_POLICY)
        csp['script-src'] = "'self' 'unsafe-inline'"
        csp['style-src'] = "'self' 'unsafe-inline'"
        self.headers['Content-Security-Policy'] = csp.get_header_value()

    def get_context_data(self, *args, **kwargs):
        """Add user's TURN server information to view context."""
        app = app_module.App.get('janus')
        config = app.get_component('turn-janus').get_configuration()
        context = super().get_context_data(*args, **kwargs)
        context['user_turn_config'] = config.to_json()
        return context

    def get(self, request, *args, **kwargs):
        """Handle GET request and return a response object."""
        context = self.get_context_data(**kwargs)
        return self.render_to_response(context, headers=self.headers)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Application manifest for security.
"""

from django.utils.translation import gettext_lazy as _

backup = {'config': {'files': ['/etc/security/access.d/50freedombox.conf']}}

tags = [_('Automatic bans'), _('Reports')]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
URLs for the Users module
"""

from django.urls import re_path
from stronghold.decorators import public

from plinth.utils import non_admin_view

from . import views

urlpatterns = [
    re_path(r'^sys/users/$', views.UserList.as_view(), name='index'),
    re_path(r'^sys/users/create/$', views.UserCreate.as_view(), name='create'),
    re_path(r'^sys/users/(?P<slug>[\w.@+-]+)/edit/$',
            non_admin_view(views.UserUpdate.as_view()), name='edit'),
    re_path(r'^sys/users/(?P<slug>[\w.@+-]+)/change_password/$',
            non_admin_view(views.UserChangePassword.as_view()),
            name='change_password'),
    re_path(r'^sys/users/(?P<username>[\w.@+-]+)/passkeys/$',
            non_admin_view(views.PasskeysList.as_view()), name='passkeys'),
    re_path(r'^sys/users/(?P<username>[\w.@+-]+)/passkeys/add-begin/$',
            non_admin_view(views.passkey_add_begin), name='passkey_add_begin'),
    re_path(r'^sys/users/(?P<username>[\w.@+-]+)/passkeys/add-complete/$',
            non_admin_view(views.passkey_add_complete),
            name='passkey_add_complete'),
    re_path(
        r'^sys/users/(?P<username>[\w.@+-]+)/passkeys/'
        r'(?P<passkey_id>[\d]+)/edit/$',
        non_admin_view(views.PasskeyEdit.as_view()), name='passkey_edit'),
    re_path(r'^accounts/login/passkey-begin/$',
            public(views.passkey_login_begin), name='passkey_login_begin'),
    re_path(r'^accounts/login/passkey-complete/$',
            public(views.passkey_login_complete),
            name='passkey_login_complete'),
    re_path(
        r'^sys/users/(?P<username>[\w.@+-]+)/passkeys/'
        r'(?P<passkey_id>[\d]+)/delete/$',
        non_admin_view(views.PasskeyDelete.as_view()), name='passkey_delete'),
    re_path(r'^accounts/login/$', public(views.LoginView.as_view()),
            name='login'),
    re_path(r'^accounts/logout/$', public(views.logout), name='logout'),
    re_path(r'^users/firstboot/$', public(views.FirstBootView.as_view()),
            name='firstboot'),
    re_path(r'accounts/login/locked/$', public(views.CaptchaView.as_view()),
            name='locked_out'),
]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Application manifest for Nextcloud."""

from django.utils.translation import gettext_lazy as _

from plinth.clients import store_url

_nextcloud_android_package_id = 'com.nextcloud.client'

clients = [{
    'name': _('Nextcloud'),
    'platforms': [{
        'type': 'web',
        'url': '/nextcloud/'
    }]
}, {
    'name':
        _('Nextcloud'),
    'platforms': [{
        'type': 'download',
        'os': 'gnu-linux',
        'url': 'https://nextcloud.com/install/#install-clients'
    }, {
        'type': 'download',
        'os': 'macos',
        'url': 'https://nextcloud.com/install/#install-clients'
    }, {
        'type': 'download',
        'os': 'windows',
        'url': 'https://nextcloud.com/install/#install-clients'
    }, {
        'type': 'store',
        'os': 'ios',
        'store_name': 'app-store',
        'url': 'https://itunes.apple.com/us/app/nextcloud/id1125420102?mt=8'
    }, {
        'type': 'store',
        'os': 'android',
        'store_name': 'google-play',
        'url': store_url('google-play', _nextcloud_android_package_id)
    }, {
        'type': 'store',
        'os': 'android',
        'store_name': 'f-droid',
        'url': store_url('f-droid', _nextcloud_android_package_id)
    }]
}]

backup = {
    'data': {
        'directories': ['/var/lib/nextcloud/'],
        'files': ['/var/lib/plinth/backups-data/nextcloud-database.sql']
    }
}

tags = [_('File sync'), _('Sharing'), _('Groupware'), _('Non-Debian')]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Configure ikiwiki."""

import os
import pathlib
import re
import shutil

from plinth import action_utils
from plinth.actions import privileged, secret_str

SETUP_WIKI = '/etc/ikiwiki/plinth-wiki.setup'
SETUP_BLOG = '/etc/ikiwiki/plinth-blog.setup'
SITE_PATH = '/var/www/ikiwiki'
WIKI_PATH = '/var/lib/ikiwiki'


def _is_safe_path(basedir, path):
    """Return whether a path is safe."""
    return os.path.realpath(path).startswith(basedir)


@privileged
def setup():
    """Write Apache configuration and wiki/blog setup scripts."""
    if not os.path.exists(SITE_PATH):
        os.makedirs(SITE_PATH)


def _get_title(site):
    """Get blog or wiki title."""
    try:
        with open(os.path.join(SITE_PATH, site, 'index.html'),
                  encoding='utf-8') as index_file:
            match = re.search(r'<title>(.*)</title>', index_file.read())
            if match:
                return match[1]
    except FileNotFoundError:
        pass

    return site


@privileged
def get_sites() -> list[tuple[str, str]]:
    """Get wikis and blogs."""
    sites = []
    if os.path.exists(SITE_PATH):
        for site in os.listdir(SITE_PATH):
            if not os.path.isdir(os.path.join(SITE_PATH, site)):
                continue

            title = _get_title(site)
            sites.append((site, title))

    return sites


@privileged
def create_wiki(wiki_name: str, admin_name: str, admin_password: secret_str):
    """Create a wiki."""
    pw_bytes = admin_password.encode()
    input_ = pw_bytes + b'\n' + pw_bytes
    action_utils.run(['ikiwiki', '-setup', SETUP_WIKI, wiki_name, admin_name],
                     input=input_, env=dict(os.environ,
                                            PERL_UNICODE='AS'), check=True)


@privileged
def create_blog(blog_name: str, admin_name: str, admin_password: secret_str):
    """Create a blog."""
    pw_bytes = admin_password.encode()
    input_ = pw_bytes + b'\n' + pw_bytes
    action_utils.run(['ikiwiki', '-setup', SETUP_BLOG, blog_name, admin_name],
                     input=input_, env=dict(os.environ, PERL_UNICODE='AS'))


@privileged
def setup_site(site_name: str):
    """Run setup for a site."""
    setup_path = os.path.join(WIKI_PATH, site_name + '.setup')
    action_utils.run(['ikiwiki', '-setup', setup_path], check=True)


@privileged
def delete(name: str):
    """Delete a wiki or blog."""
    html_folder = os.path.join(SITE_PATH, name)
    wiki_folder = os.path.join(WIKI_PATH, name)

    if not (_is_safe_path(SITE_PATH, html_folder)
            and _is_safe_path(WIKI_PATH, wiki_folder)):
        raise ValueError(
            'Error: {0} is not a correct wiki/blog name.'.format(name))

    shutil.rmtree(html_folder, ignore_errors=True)
    shutil.rmtree(wiki_folder, ignore_errors=True)
    shutil.rmtree(wiki_folder + '.git', ignore_errors=True)
    pathlib.Path(wiki_folder + '.setup').unlink(missing_ok=True)


@privileged
def uninstall():
    """Remove all ikiwiki sites."""
    for site in get_sites():
        delete(site[0])



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Functional, browser based tests for users app.
"""

# TODO Scenario: Add user to wiki group
# TODO Scenario: Remove user from wiki group

import subprocess
import urllib

import pytest

from plinth.tests import functional

_admin_password = functional.config['DEFAULT']['password']

pytestmark = [pytest.mark.system, pytest.mark.essential, pytest.mark.users]

_language_codes = {
    'None': '',
    'Deutsch': 'de',
    'Nederlands': 'nl',
    'Português': 'pt',
    'Türkçe': 'tr',
    'dansk': 'da',
    'español': 'es',
    'français': 'fr',
    'norsk (bokmål)': 'nb',
    'polski': 'pl',
    'svenska': 'sv',
    'Русский': 'ru',
    'తెలుగు': 'te',
    '简体中文': 'zh-hans'
}

_config_page_title_language_map = {
    '': 'General Configuration',
    'da': 'Generel Konfiguration',
    'de': 'Allgemeine Konfiguration',
    'es': 'Configuración general',
    'fr': 'Configuration générale',
    'nb': 'Generelt oppsett',
    'nl': 'Algemene Instellingen',
    'pl': 'Ustawienia główne',
    'pt': 'Configuração Geral',
    'ru': 'Общие настройки',
    'sv': 'Allmän Konfiguration',
    'te': 'సాధారణ ఆకృతీకరణ',
    'tr': 'Genel Yapılandırma',
    'zh-hans': '常规配置',
}


@pytest.fixture(scope='module', autouse=True)
def fixture_background(session_browser):
    """Unset language."""
    yield
    functional.login(session_browser)
    functional.user_set_language(session_browser, _language_codes['None'])


@pytest.fixture(scope='function', autouse=True)
def fixture_login(session_browser):
    """Login."""
    functional.login(session_browser)


@pytest.fixture(name='syncthing_installed')
def fixture_syncthing_installed(session_browser):
    """Login and install the app."""
    functional.login(session_browser)
    functional.install(session_browser, 'syncthing')
    functional.app_enable(session_browser, 'syncthing')
    yield
    functional.app_disable(session_browser, 'syncthing')


def test_app_access(session_browser, syncthing_installed):
    """Test that only logged-in users can access Syncthing web interface."""
    functional.logout(session_browser)
    functional.access_url(session_browser, 'syncthing')
    assert functional.is_login_prompt(session_browser)

    functional.login(session_browser)
    functional.access_url(session_browser, 'syncthing')
    assert functional.is_available(session_browser, 'syncthing')


def test_create_user(session_browser):
    """Test creating a user."""
    _delete_user(session_browser, 'alice')

    functional.create_user(session_browser, 'alice', email='alice@example.com')
    assert functional.user_exists(session_browser, 'alice')
    assert _get_email(session_browser, 'alice') == 'alice@example.com'


def test_rename_user(session_browser, host_sudo):
    """Test renaming a user."""
    _non_admin_user_exists(session_browser, 'alice')
    _delete_user(session_browser, 'bob')

    _rename_user(session_browser, 'alice', 'bob')
    assert not functional.user_exists(session_browser, 'alice')
    assert functional.user_exists(session_browser, 'bob')

    assert 'cn: bob' in host_sudo.check_output('ldapfinger bob')
    assert host_sudo.user('bob').gecos == 'bob'
    assert host_sudo.user('bob').home == '/home/bob'


def test_admin_users_can_change_own_ssh_keys(session_browser):
    """Test that admin users can change their own ssh keys."""
    _set_ssh_keys(session_browser, 'somekey123')
    assert _get_ssh_keys(session_browser) == 'somekey123'


def test_non_admin_users_can_change_own_ssh_keys(session_browser):
    """Test that non-admin users can change their own ssh keys."""
    _non_admin_user_exists(session_browser, 'alice')
    functional.login_with_account(session_browser, functional.base_url,
                                  'alice')
    _set_ssh_keys(session_browser, 'somekey456')
    assert _get_ssh_keys(session_browser) == 'somekey456'


def test_admin_users_can_change_other_users_ssh_keys(session_browser):
    """Test that admin users can change other user's ssh keys."""
    _non_admin_user_exists(session_browser, 'alice')
    _set_ssh_keys(session_browser, 'alicesomekey123', username='alice')
    assert _get_ssh_keys(session_browser,
                         username='alice') == 'alicesomekey123'


def test_users_can_remove_ssh_keys(session_browser):
    """Test that users can remove ssh keys."""
    _set_ssh_keys(session_browser, 'somekey123')
    _set_ssh_keys(session_browser, '')
    assert _get_ssh_keys(session_browser) == ''


def test_users_can_connect_passwordless_over_ssh(session_browser,
                                                 tmp_path_factory):
    """Test that users can connect passwordless over ssh if the keys are
    set."""
    functional.app_enable(session_browser, 'ssh')
    _configure_ssh_keys(session_browser, tmp_path_factory)
    _should_connect_passwordless_over_ssh(session_browser, tmp_path_factory)


def test_ssh_passwordless_after_user_rename(session_browser, tmp_path_factory):
    """Test that users can connect passwordless after user is renamed."""
    username_old = 'bob'
    username_new = 'bob2'
    functional.app_enable(session_browser, 'ssh')
    _non_admin_user_exists(session_browser, username_old,
                           groups=['freedombox-ssh'])
    _delete_user(session_browser, username_new)
    _configure_ssh_keys(session_browser, tmp_path_factory,
                        username=username_old)
    _should_connect_passwordless_over_ssh(session_browser, tmp_path_factory,
                                          username=username_old)

    _rename_user(session_browser, username_old, username_new)

    assert functional.user_exists(session_browser, username_new)
    _should_connect_passwordless_over_ssh(session_browser, tmp_path_factory,
                                          username=username_new)


def test_users_cannot_connect_passwordless_over_ssh(session_browser,
                                                    tmp_path_factory):
    """Test that users cannot connect passwordless over ssh if the keys aren't
    set."""
    functional.app_enable(session_browser, 'ssh')
    _configure_ssh_keys(session_browser, tmp_path_factory)
    _set_ssh_keys(session_browser, '')
    _should_not_connect_passwordless_over_ssh(session_browser,
                                              tmp_path_factory)


def test_update_user(session_browser):
    """Test changing properties of a user."""
    functional.create_user(session_browser, 'alice', email='alice@example.com')

    # Update email
    _set_email(session_browser, 'alice', 'alice1@example.com')
    assert _get_email(session_browser, 'alice') == 'alice1@example.com'
    _set_email(session_browser, 'alice', 'alice2@example.com')
    assert _get_email(session_browser, 'alice') == 'alice2@example.com'


@pytest.mark.parametrize('language_code', _language_codes.values())
def test_change_language(session_browser, language_code):
    """Test changing the language."""
    functional.user_set_language(session_browser, language_code)
    assert _check_language(session_browser, language_code)


def test_user_states(session_browser, tmp_path_factory):
    """Test that admin users can set other users as inactive/active."""
    username = 'bob2'
    _non_admin_user_exists(session_browser, username,
                           groups=['freedombox-ssh'])
    _configure_ssh_keys(session_browser, tmp_path_factory, username=username)

    # Test set user inactive
    _set_user_status(session_browser, username, 'inactive')
    # Test Django login
    _cannot_log_in(session_browser, username)
    # Test PAM/nslcd authorization
    _should_not_connect_passwordless_over_ssh(session_browser,
                                              tmp_path_factory,
                                              username=username)

    # Test set user active
    functional.login(session_browser)
    _set_user_status(session_browser, username, 'active')
    _can_log_in(session_browser, username)
    _should_connect_passwordless_over_ssh(session_browser, tmp_path_factory,
                                          username=username)


def test_admin_users_can_change_own_password(session_browser):
    """Test that admin users can change their own password."""
    _admin_user_exists(session_browser, 'testadmin')
    functional.login_with_account(session_browser, functional.base_url,
                                  'testadmin')
    _change_password(session_browser, 'newpassword456')
    _can_log_in_with_password(session_browser, 'testadmin', 'newpassword456')


def test_admin_users_can_change_others_password(session_browser):
    """Test that admin users can change other user's password."""
    _non_admin_user_exists(session_browser, 'alice')
    _change_password(session_browser, 'secretsecret567', username='alice')
    _can_log_in_with_password(session_browser, 'alice', 'secretsecret567')


def test_non_admin_users_can_change_own_password(session_browser):
    """Test that non-admin users can change their own password."""
    _non_admin_user_exists(session_browser, 'alice')
    functional.login_with_account(session_browser, functional.base_url,
                                  'alice')
    _change_password(session_browser, 'newpassword123')
    _can_log_in_with_password(session_browser, 'alice', 'newpassword123')


def test_delete_user(session_browser):
    """Test deleting a user."""
    _non_admin_user_exists(session_browser, 'alice')
    functional.delete_user(session_browser, 'alice')
    assert not functional.user_exists(session_browser, 'alice')


def _delete_user(session_browser, name):
    """Delete a user."""
    if functional.user_exists(session_browser, name):
        functional.delete_user(session_browser, name)


def _admin_user_exists(session_browser, name):
    _delete_user(session_browser, name)
    functional.create_user(session_browser, name, groups=['admin'])


def _non_admin_user_exists(session_browser, name, groups=[]):
    _delete_user(session_browser, name)
    functional.create_user(session_browser, name, groups=groups)


def _generate_ssh_keys(session_browser, key_file):
    try:
        key_file.unlink()
    except FileNotFoundError:
        pass

    subprocess.check_call(
        ['ssh-keygen', '-t', 'ed25519', '-N', '', '-q', '-f',
         str(key_file)])


def _configure_ssh_keys(session_browser, tmp_path_factory, username=None):
    key_file = tmp_path_factory.getbasetemp() / 'users-ssh.key'
    _generate_ssh_keys(session_browser, key_file)
    public_key_file = key_file.with_suffix(key_file.suffix + '.pub')
    public_key = public_key_file.read_text()
    _set_ssh_keys(session_browser, public_key, username=username)


def _can_log_in(session_browser, username):
    functional.login_with_account(session_browser, functional.base_url,
                                  username)
    assert len(session_browser.find_by_id('id_user_menu')) > 0


def _can_log_in_with_password(session_browser, username, password):
    functional.logout(session_browser)
    functional.login_with_account(session_browser, functional.base_url,
                                  username, password)
    assert len(session_browser.find_by_id('id_user_menu')) > 0


def _cannot_log_in(session_browser, username):
    functional.login_with_account(session_browser, functional.base_url,
                                  username)
    assert len(session_browser.find_by_id('id_user_menu')) == 0


def _should_connect_passwordless_over_ssh(session_browser, tmp_path_factory,
                                          username=None):
    key_file = tmp_path_factory.getbasetemp() / 'users-ssh.key'
    _try_login_to_ssh(key_file=key_file, username=username)


def _should_not_connect_passwordless_over_ssh(session_browser,
                                              tmp_path_factory, username=None):
    key_file = tmp_path_factory.getbasetemp() / 'users-ssh.key'
    with pytest.raises(subprocess.CalledProcessError):
        _try_login_to_ssh(key_file=key_file, username=username)


def _rename_user(browser, old_name, new_name):
    functional.nav_to_module(browser, 'users')
    functional.click_link_by_href(browser,
                                  f'/freedombox/sys/users/{old_name}/edit/')
    browser.find_by_id('id_username').fill(new_name)
    browser.find_by_id('id_confirm_password').fill(_admin_password)
    functional.submit(browser, form_class='form-update')


def _set_email(browser, username, email):
    """Set the email field value for a user."""
    functional.visit(browser,
                     '/freedombox/sys/users/{}/edit/'.format(username))
    browser.find_by_id('id_email').fill(email)
    browser.find_by_id('id_confirm_password').fill(_admin_password)
    functional.submit(browser, form_class='form-update')


def _get_email(browser, username):
    """Return the email field value for a user."""
    functional.visit(browser,
                     '/freedombox/sys/users/{}/edit/'.format(username))
    return browser.find_by_id('id_email').value


def _check_language(browser, language_code):
    functional.nav_to_module(browser, 'config')
    return browser.find_by_css('.app-titles').first.find_by_tag(
        'h2').first.value == _config_page_title_language_map[language_code]


def _get_ssh_keys(browser, username=None):
    functional.visit(browser, '/freedombox/')
    if username is None:
        browser.find_by_id('id_user_menu').click()
        functional.click_and_wait(browser,
                                  browser.find_by_id('id_user_edit_menu'))
    else:
        functional.visit(browser, f'/freedombox/sys/users/{username}/edit/')
    return browser.find_by_id('id_ssh_keys').text


def _set_ssh_keys(browser, ssh_keys, username=None):
    if username is None:
        browser.find_by_id('id_user_menu').click()
        functional.click_and_wait(browser,
                                  browser.find_by_id('id_user_edit_menu'))
    else:
        functional.visit(browser, f'/freedombox/sys/users/{username}/edit/')

    current_user = browser.find_by_id('id_user_menu_link').text
    auth_password = functional.get_password(current_user)

    browser.find_by_id('id_ssh_keys').fill(ssh_keys)
    browser.find_by_id('id_confirm_password').fill(auth_password)

    functional.submit(browser, form_class='form-update')


def _set_user_status(browser, username, status):
    functional.visit(browser,
                     '/freedombox/sys/users/{}/edit/'.format(username))
    if status == 'inactive':
        browser.find_by_id('id_is_active').uncheck()
    elif status == 'active':
        browser.find_by_id('id_is_active').check()

    browser.find_by_id('id_confirm_password').fill(_admin_password)
    functional.submit(browser, form_class='form-update')


def _change_password(browser, new_password, current_password=None,
                     username=None):
    if username is None:
        browser.find_by_id('id_user_menu').click()
        functional.click_and_wait(
            browser, browser.find_by_id('id_change_password_menu'))
    else:
        functional.visit(
            browser,
            '/freedombox/sys/users/{}/change_password/'.format(username))

    current_user = browser.find_by_id('id_user_menu_link').text
    auth_password = current_password or functional.get_password(current_user)

    browser.find_by_id('id_new_password1').fill(new_password)
    browser.find_by_id('id_new_password2').fill(new_password)
    browser.find_by_id('id_confirm_password').fill(auth_password)
    functional.submit(browser, form_class='form-change-password')


def _try_login_to_ssh(key_file=None, username=None):
    user = username if username else functional.config['DEFAULT']['username']
    hostname = urllib.parse.urlparse(
        functional.config['DEFAULT']['url']).hostname
    port = functional.config['DEFAULT']['ssh_port']

    subprocess.check_call([
        'ssh', '-p', port, '-i', key_file, '-q', '-o',
        'StrictHostKeyChecking=no', '-o', 'UserKnownHostsFile=/dev/null', '-o',
        'BatchMode=yes', '{0}@{1}'.format(user, hostname), '/bin/true'
    ])



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Configure Shadowsocks."""

import json
import os
import pathlib
import random
import string
from shutil import move

from plinth import action_utils
from plinth.actions import privileged, secret_str

SHADOWSOCKS_CONFIG_SYMLINK = '/etc/shadowsocks-libev/freedombox.json'
SHADOWSOCKS_CONFIG_ACTUAL = \
    '/var/lib/private/shadowsocks-libev/freedombox/freedombox.json'


@privileged
def setup():
    """Perform initial setup steps."""
    # Only client socks5 proxy is supported for now. Disable the
    # server component.
    action_utils.service_disable('shadowsocks-libev')

    os.makedirs('/var/lib/private/shadowsocks-libev/freedombox/',
                exist_ok=True)

    # if existing configuration from version 1 which is normal file
    # move it to new location.
    if (not os.path.islink(SHADOWSOCKS_CONFIG_SYMLINK)
            and os.path.isfile(SHADOWSOCKS_CONFIG_SYMLINK)):
        move(SHADOWSOCKS_CONFIG_SYMLINK, SHADOWSOCKS_CONFIG_ACTUAL)

    if not os.path.islink(SHADOWSOCKS_CONFIG_SYMLINK):
        os.symlink(SHADOWSOCKS_CONFIG_ACTUAL, SHADOWSOCKS_CONFIG_SYMLINK)

    if not os.path.isfile(SHADOWSOCKS_CONFIG_ACTUAL):
        password = ''.join(
            random.choice(string.ascii_letters) for _ in range(12))
        initial_config = {
            'server': '127.0.0.1',
            'server_port': 8388,
            'local_port': 1080,
            'password': password,
            'method': 'chacha20-ietf-poly1305'
        }
        _merge_config(initial_config)

    # Commit 50e5608331330b37c0b9cce846e34ccc193d1b0d incorrectly sets the
    # StateDirectory without setting DynamicUser. Buster's shadowsocks will
    # then create directory /var/lib/shadowsocks-libev/freedombox/ and refuse
    # to delete is in later versions when DynamicUser=yes needs it to be a
    # symlink.
    action_utils.service_daemon_reload()
    wrong_state_dir = pathlib.Path('/var/lib/shadowsocks-libev/freedombox/')
    if not wrong_state_dir.is_symlink() and wrong_state_dir.is_dir():
        wrong_state_dir.rmdir()

    from plinth.modules.shadowsocks import ShadowsocksApp
    if action_utils.service_is_enabled(ShadowsocksApp.DAEMON):
        action_utils.service_restart(ShadowsocksApp.DAEMON)


@privileged
def get_config() -> dict[str, int | str]:
    """Read and print Shadowsocks configuration."""
    config = open(SHADOWSOCKS_CONFIG_SYMLINK, 'r', encoding='utf-8').read()
    return json.loads(config)


def _merge_config(config):
    """Write merged configuration into file."""
    try:
        current_config = open(SHADOWSOCKS_CONFIG_SYMLINK, 'r',
                              encoding='utf-8').read()
        current_config = json.loads(current_config)
    except (OSError, json.JSONDecodeError):
        current_config = {}

    new_config = current_config
    new_config.update(config)
    new_config = json.dumps(new_config, indent=4, sort_keys=True)
    open(SHADOWSOCKS_CONFIG_SYMLINK, 'w', encoding='utf-8').write(new_config)


@privileged
def merge_config(config: dict[str, int | str], password: secret_str):
    """Configure Shadowsocks Client."""
    config['password'] = password
    _merge_config(config)

    # Don't try_restart because initial configuration may not be valid so
    # shadowsocks will not be running even when enabled.
    from . import ShadowsocksApp
    if action_utils.service_is_enabled(ShadowsocksApp.DAEMON):
        action_utils.service_restart(ShadowsocksApp.DAEMON)


@privileged
def uninstall():
    """Remove configuration files."""
    for path in SHADOWSOCKS_CONFIG_SYMLINK, SHADOWSOCKS_CONFIG_ACTUAL:
        pathlib.Path(path).unlink(missing_ok=True)



# SPDX-License-Identifier: AGPL-3.0-or-later

from django.utils.translation import gettext_lazy as _

clients = [{
    'name': _('calibre'),
    'platforms': [{
        'type': 'web',
        'url': '/calibre/'
    }]
}]

backup = {
    'data': {
        'directories': ['/var/lib/private/calibre-server-freedombox/']
    },
    'services': ['calibre-server-freedombox'],
    'delete_before_restore': ['/var/lib/private/calibre-server-freedombox/']
}

tags = [_('Ebook'), _('Library'), _('Ebook reader')]



# SPDX-License-Identifier: AGPL-3.0-or-later

from django.utils.translation import gettext_lazy as _

from plinth.clients import store_url

clients = [{
    'name':
        _('Luanti'),
    'platforms': [{
        'type': 'download',
        'os': 'gnu-linux',
        'url': 'https://www.luanti.org/downloads/'
    }, {
        'type': 'download',
        'os': 'macos',
        'url': 'https://www.luanti.org/downloads/'
    }, {
        'type': 'download',
        'os': 'windows',
        'url': 'https://www.luanti.org/downloads/'
    }, {
        'type': 'store',
        'os': 'android',
        'store_name': 'f-droid',
        'url': store_url('f-droid', 'net.minetest.minetest')
    }, {
        'type': 'store',
        'os': 'android',
        'store_name': 'google-play',
        'url': store_url('google-play', 'net.minetest.minetest')
    }, {
        'type': 'package',
        'format': 'deb',
        'name': 'luanti'
    }]
}]

backup = {
    'config': {
        'files': ['/etc/luanti/default.conf']
    },
    'data': {
        'directories': ['/var/lib/private/luanti/default/']
    },
    'services': ['luanti-server']
}

tags = [_('Game server'), _('Block sandbox'), _('Platform')]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""FreedomBox app for calibre e-book library."""

import re

from django.utils.translation import gettext_lazy as _

from plinth import app as app_module
from plinth import cfg, frontpage, menu
from plinth.config import DropinConfigs
from plinth.daemon import Daemon
from plinth.modules.apache.components import Webserver
from plinth.modules.backups.components import BackupRestore
from plinth.modules.firewall.components import (Firewall,
                                                FirewallLocalProtection)
from plinth.modules.users.components import UsersAndGroups
from plinth.package import Packages
from plinth.utils import format_lazy

from . import manifest, privileged

_description = [
    format_lazy(
        _('calibre server provides online access to your e-book collection. '
          'You can store your e-books on your {box_name}, read them online or '
          'from any of your devices.'), box_name=_(cfg.box_name)),
    _('You can organize your e-books, extract and edit their metadata, and '
      'perform advanced search. calibre can import, export, or convert across '
      'a wide range of formats to make e-books ready for reading on any '
      'device. It also provides an online web reader. It remembers your '
      'last read location, bookmarks, and highlighted text. Content '
      'distribution using OPDS is currently not supported.'),
    _('Only users belonging to <em>calibre</em> group will be able to access '
      'the app. All users with access can use all the libraries.')
]

LIBRARY_NAME_PATTERN = r'[a-zA-Z0-9 _-]+'


class CalibreApp(app_module.App):
    """FreedomBox app for calibre."""

    app_id = 'calibre'

    _version = 3

    DAEMON = 'calibre-server-freedombox'

    def __init__(self) -> None:
        """Create components for the app."""
        super().__init__()

        groups = {'calibre': _('Use calibre e-book libraries')}

        info = app_module.Info(app_id=self.app_id, version=self._version,
                               name=_('calibre'), icon_filename='calibre',
                               description=_description, manual_page='Calibre',
                               clients=manifest.clients, tags=manifest.tags,
                               donation_url='https://calibre-ebook.com/donate')
        self.add(info)

        menu_item = menu.Menu('menu-calibre', info.name, info.icon_filename,
                              info.tags, 'calibre:index',
                              parent_url_name='apps')
        self.add(menu_item)

        shortcut = frontpage.Shortcut('shortcut-calibre', info.name,
                                      icon=info.icon_filename, url='/calibre',
                                      clients=info.clients, tags=info.tags,
                                      login_required=True,
                                      allowed_groups=list(groups))
        self.add(shortcut)

        packages = Packages('packages-calibre', ['calibre'])
        self.add(packages)

        dropin_configs = DropinConfigs('dropin-configs-calibre', [
            '/etc/apache2/conf-available/calibre-freedombox.conf',
        ])
        self.add(dropin_configs)

        firewall = Firewall('firewall-calibre', info.name,
                            ports=['http', 'https'], is_external=True)
        self.add(firewall)

        firewall_local_protection = FirewallLocalProtection(
            'firewall-local-protection-calibre', ['8844'])
        self.add(firewall_local_protection)

        webserver = Webserver('webserver-calibre', 'calibre-freedombox',
                              urls=['https://{host}/calibre'])
        self.add(webserver)

        daemon = Daemon('daemon-calibre', self.DAEMON,
                        listen_ports=[(8844, 'tcp4')])
        self.add(daemon)

        users_and_groups = UsersAndGroups('users-and-groups-calibre',
                                          reserved_usernames=['calibre'],
                                          groups=groups)
        self.add(users_and_groups)

        backup_restore = BackupRestore('backup-restore-calibre',
                                       **manifest.backup)
        self.add(backup_restore)

    def setup(self, old_version):
        """Install and configure the app."""
        super().setup(old_version)
        if not old_version:
            self.enable()

    def uninstall(self):
        """De-configure and uninstall the app."""
        super().uninstall()
        privileged.uninstall()


def validate_library_name(library_name):
    """Raise exception if library name does not fit the accepted pattern."""
    if not re.fullmatch(r'[A-Za-z0-9_.-]+', library_name):
        raise Exception('Invalid library name')



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Configure or run unattended-upgrades."""

import logging
import os
import pathlib
import re
import subprocess

from plinth import action_utils
from plinth.action_utils import (apt_hold_flag, apt_unhold_freedombox,
                                 is_package_manager_busy, run_apt_command,
                                 service_is_running)
from plinth.actions import privileged

from . import distupgrade, utils

logger = logging.getLogger(__name__)

BACKPORTS_SOURCES_LIST = '/etc/apt/sources.list.d/freedombox2.list'
UNSTABLE_SOURCES_LIST = pathlib.Path(
    '/etc/apt/sources.list.d/freedombox-unstable.list')

UNSTABLE_PREFERENCES = pathlib.Path(
    '/etc/apt/preferences.d/50freedombox-unstable.pref')

AUTO_CONF_FILE = '/etc/apt/apt.conf.d/20auto-upgrades'
LOG_FILE = '/var/log/unattended-upgrades/unattended-upgrades.log'
DPKG_LOG_FILE = '/var/log/unattended-upgrades/unattended-upgrades-dpkg.log'

APT_PREFERENCES_FREEDOMBOX = \
    '''Explanation: This file is managed by FreedomBox, do not edit.
Explanation: Allow carefully selected updates to 'freedombox' from backports.
Package: src:freedombox
Pin: release n={}-backports
Pin-Priority: 500
'''

# Whenever these preferences needs to change, increment the version number
# upgrades app. This ensures that setup is run again and the new contents are
# overwritten on the old file.
APT_PREFERENCES_APPS = \
    '''Explanation: This file is managed by FreedomBox, do not edit.
Explanation: matrix-synapse shall not be available in Debian stable but
Explanation: only in backports. Upgrade priority of packages that have needed
Explanation: versions only in backports.
Explanation: matrix-synapse >= 1.92.0-3 requires
Explanation: python3-canonicaljson >= 2.0.0~
Package: python3-canonicaljson
Pin: release n=bookworm-backports
Pin-Priority: 500

Explanation: Prevent installation of Samba Active Directory. AD package depends
Explanation: on winbind, which breaks FreedomBox LDAP PAM configuration.
Explanation: In Debian Trixie, AD server package is required by samba package,
Explanation: but is not required to run Samba file server. See also Debian
Explanation: bug report 1099755.
Package: samba-ad-dc
Pin: release *
Pin-Priority: -1

Explanation: Make matrix-synapse package and its dependencies installable from
Explanation: Debian 'unstable' distribution.
Package: matrix-synapse
Pin: release n=sid
Pin-Priority: 200

Explanation: matrix-synapse depends on python3-python-multipart
Package: python3-python-multipart
Pin: release n=sid
Pin-Priority: 200

Explanation: matrix-synapse recommends python3-pympler
Package: python3-pympler
Pin: release n=sid
Pin-Priority: 200

Explanation: Make janus package and its dependencies installable from Debian
Explanation: 'unstable' distribution.
Package: janus
Pin: release n=sid
Pin-Priority: 200

Explanation: Janus app in FreedomBox needs the package for web UI.
Package: libjs-janus-gateway
Pin: release n=sid
Pin-Priority: 200
'''

APT_PREFERENCES_UNSTABLE = \
    '''Explanation: This file is managed by FreedomBox, do not edit.
Explanation: De-prioritize all the packages from Unstable distribution.
Explanation: The priority of packages in *-backports will be set to 300.
Explanation: Prioritize unstable lower than packages in backports.
Package: *
Pin: release n=sid
Pin-Priority: -100

Explanation: The priority of packages in *-backports will be 100 by default.
Explanation: Prioritize them higher than unstable packages.
Package: *
Pin: release n=trixie-backports
Pin-Priority: 300

Explanation: The priority of packages in *-backports will be 100 by default.
Explanation: Prioritize them higher than unstable packages.
Package: *
Pin: release n=bookworm-backports
Pin-Priority: 300
'''

APT_UNSTABLE_SOURCES = \
    '''# This file is managed by FreedomBox, do not edit.
# Allow carefully selected updates to 'freedombox' from unstable.

deb {protocol}://deb.debian.org/debian unstable main
deb-src {protocol}://deb.debian.org/debian unstable main
'''


def _release_held_freedombox():
    """If freedombox package was left in held state, release it.

    This can happen due to an interrupted process.
    """
    if apt_hold_flag.exists() and not is_package_manager_busy():
        apt_unhold_freedombox()


@privileged
def release_held_packages():
    """Release any packages that are being held."""
    if is_package_manager_busy():
        logger.warning('Package manager is busy, skipping releasing holds.')
        return

    if service_is_running('freedombox-dist-upgrade'):
        logger.warning('Distribution upgrade in progress, skipping releasing '
                       'holds.')
        return

    output = action_utils.run(['apt-mark', 'showhold'],
                              check=True).stdout.decode().strip()
    holds = output.split('\n')
    logger.info('Releasing package holds: %s', holds)
    action_utils.run(['apt-mark', 'unhold', *holds], check=True)


@privileged
def run():
    """Run unattended-upgrades."""
    action_utils.run(['dpkg', '--configure', '-a'], check=False)
    run_apt_command(['--fix-broken', 'install'])
    _release_held_freedombox()

    subprocess.Popen(['systemctl', 'start', 'freedombox-manual-upgrade'],
                     stdin=subprocess.DEVNULL, stdout=subprocess.DEVNULL,
                     stderr=subprocess.DEVNULL, close_fds=True,
                     start_new_session=True)


@privileged
def check_auto() -> bool:
    """Check if automatic upgrades are enabled."""
    return utils.check_auto()


@privileged
def enable_auto():
    """Enable automatic upgrades."""
    with open(AUTO_CONF_FILE, 'w', encoding='utf-8') as conffile:
        conffile.write('APT::Periodic::Update-Package-Lists "1";\n')
        conffile.write('APT::Periodic::Unattended-Upgrade "1";\n')


@privileged
def disable_auto():
    """Disable automatic upgrades."""
    with open(AUTO_CONF_FILE, 'w', encoding='utf-8') as conffile:
        conffile.write('APT::Periodic::Update-Package-Lists "0";\n')
        conffile.write('APT::Periodic::Unattended-Upgrade "0";\n')


@privileged
def get_log() -> str:
    """Return the automatic upgrades log."""
    log_lines = []
    try:
        log_lines.append('==> ' + os.path.basename(LOG_FILE))
        with open(LOG_FILE, 'r', encoding='utf-8') as file_handle:
            log_lines.append(file_handle.read())
    except IOError:
        pass

    try:
        log_lines.append('==> ' + os.path.basename(DPKG_LOG_FILE))
        with open(DPKG_LOG_FILE, 'r', encoding='utf-8') as file_handle:
            log_lines.append(file_handle.read())
    except IOError:
        pass

    return '\n'.join(log_lines)


def _add_backports_sources(sources_list: str, protocol: str, dist: str):
    """Add backports sources to freedombox repositories list."""
    sources = '''# This file is managed by FreedomBox, do not edit.
# Allow carefully selected updates to 'freedombox' from backports.

deb {protocol}://deb.debian.org/debian {dist}-backports main
deb-src {protocol}://deb.debian.org/debian {dist}-backports main
'''
    sources = sources.format(protocol=protocol, dist=dist)
    with open(sources_list, 'w', encoding='utf-8') as file_handle:
        file_handle.write(sources)


def _check_and_backports_sources(develop=False):
    """Add backports sources after checking if it is available."""
    old_sources_list = '/etc/apt/sources.list.d/freedombox.list'
    if os.path.exists(old_sources_list):
        os.remove(old_sources_list)

    from plinth.modules.upgrades import is_backports_current
    if is_backports_current():
        logging.info('Repositories list up-to-date. Skipping update.')
        return

    try:
        with open('/etc/dpkg/origins/default', 'r',
                  encoding='utf-8') as default_origin:
            matches = [
                re.match(r'Vendor:\s+(Debian|FreedomBox)', line,
                         flags=re.IGNORECASE)
                for line in default_origin.readlines()
            ]
    except FileNotFoundError:
        logging.info('Could not open /etc/dpkg/origins/default')
        return

    if not any(matches):
        logging.info('System is running a derivative of Debian. Skip enabling '
                     'backports.')
        return

    if utils.is_distribution_rolling() and not develop:
        logging.info(
            'System release is unstable/testing. Skip enabling backports.')
        return

    protocol = utils.get_http_protocol()
    if protocol == 'tor+http':
        logging.info('Package download over Tor is enabled.')

    _, dist = utils.get_current_release()
    if not utils.is_release_file_available(protocol, dist, backports=True):
        logging.info(
            f'Release file for {dist}-backports is not available yet.')
        return

    print(f'{dist}-backports is now available. Adding to sources.')
    _add_backports_sources(BACKPORTS_SOURCES_LIST, protocol, dist)
    # In case of dist upgrade, rewrite the preferences file.
    _add_apt_preferences()


def _add_apt_preferences():
    """Setup APT preferences to upgrade selected packages from backports."""
    base_path = pathlib.Path('/etc/apt/preferences.d')
    for file_name in ['50freedombox.pref', '50freedombox2.pref']:
        full_path = base_path / file_name
        if full_path.exists():
            full_path.unlink()

    # Don't try to remove 50freedombox3.pref as this file is shipped with the
    # Debian package and is removed using maintainer scripts.

    if utils.is_distribution_unstable():
        logging.info(
            'System distribution is "unstable". Skip setting apt preferences '
            'for backports.')
    else:
        _, dist = utils.get_current_release()
        logging.info(f'Setting apt preferences for {dist}-backports.')
        with open(base_path / '50freedombox4.pref', 'w',
                  encoding='utf-8') as file_handle:
            file_handle.write(APT_PREFERENCES_FREEDOMBOX.format(dist))
        with open(base_path / '51freedombox-apps.pref', 'w',
                  encoding='utf-8') as file_handle:
            file_handle.write(APT_PREFERENCES_APPS)


@privileged
def setup():
    """Setup apt preferences."""
    _add_apt_preferences()


@privileged
def activate_backports(develop: bool = False):
    """Setup software repositories needed for FreedomBox.

    Repositories list for now only contains the backports. If the file exists,
    assume that it contains backports.
    """
    _check_and_backports_sources(develop)


@privileged
def activate_unstable():
    """Setup apt sources for unstable distribution and de-prioritize it.

    Select packages will be made installable from unstable.
    """
    # Operation already performed, don't write to files unnecessarily.
    if UNSTABLE_SOURCES_LIST.exists() and UNSTABLE_PREFERENCES.exists():
        logging.info('Skipping already added unstable sources.')
        return

    # If the distribution is already 'unstable', default sources.list already
    # contains sources for 'unstable'. Also, don't de-prioritize the primary
    # set of packages.
    if utils.is_distribution_unstable():
        logging.info(
            'Skipping adding unstable sources for unstable distribution.')
        return

    protocol = utils.get_http_protocol()
    if protocol == 'tor+http':
        logging.info('Package download over Tor is enabled.')

    logger.info('Adding unstable sources to apt.')
    sources = APT_UNSTABLE_SOURCES.format(protocol=protocol)
    UNSTABLE_SOURCES_LIST.write_text(sources)
    UNSTABLE_PREFERENCES.write_text(APT_PREFERENCES_UNSTABLE)


@privileged
def start_dist_upgrade():
    """Start dist upgrade process.

    Check if a new stable release is available, and start dist-upgrade process
    if updates are enabled.
    """
    _release_held_freedombox()

    distupgrade.start_service()


@privileged
def dist_upgrade():
    """Perform major distribution upgrade."""
    distupgrade.perform()


@privileged
def dist_upgrade_on_complete():
    """Perform cleanup operations after distribution upgrade."""
    distupgrade.on_complete()



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Tests for common FreedomBox views.
"""

import json

import pytest
from django.http import JsonResponse
from django.urls import resolve

from plinth.views import get_breadcrumbs, is_safe_url, json_exception


def test_get_breadcrumbs(rf, test_menu):
    """Test that computing breadcrumbs works."""

    def _crumb(name: str, is_active: bool = False, url_name: str | None = None,
               is_active_section: bool = False):
        crumb = {'name': name, 'is_active': is_active, 'url_name': url_name}
        if is_active_section:
            crumb['is_active_section'] = True

        return crumb

    def _get(path: str):
        request = rf.get(path)
        request.resolver_match = resolve(path)
        return get_breadcrumbs(request)

    def _compare(dict1: dict[str, dict[str, str | bool]],
                 dict2: dict[str, dict[str, str | bool]]):
        """Compare dictionaries with order."""
        assert list(dict1.items()) == list(dict2.items())

    _compare(_get('/'), {'/': _crumb('Home', True, 'index', True)})
    _compare(
        _get('/apps/'), {
            '/apps/': _crumb('Apps', True, 'apps', True),
            '/': _crumb('Home', False, 'index'),
        })
    _compare(
        _get('/apps/testapp/'), {
            '/apps/testapp/': _crumb('Test App', True, 'testapp:index'),
            '/apps/': _crumb('Apps', False, 'apps', True),
            '/': _crumb('Home', False, 'index'),
        })
    _compare(
        _get('/apps/testapp/create/'), {
            '/apps/testapp/create/': _crumb('Here', True, 'testapp:create'),
            '/apps/testapp/': _crumb('Test App', False, 'testapp:index'),
            '/apps/': _crumb('Apps', False, 'apps', True),
            '/': _crumb('Home', False, 'index'),
        })
    _compare(
        _get('/test/1/2/3/'), {
            '/test/1/2/3/': _crumb('Here', True, 'test', True),
            '/': _crumb('Home', False, 'index'),
        })


@pytest.mark.parametrize('url', [
    '/freedombox/login/',
    '/',
    'safe',
])
def test_is_safe_url_valid_url(url):
    """Test valid URLs for safe URL checks."""
    assert is_safe_url(url)


@pytest.mark.parametrize(
    'url',
    [
        '',
        None,
        '\\freedombox',
        '///freedombox',
        'https://example.com/freedombox/login/',
        'https:///example.com',
        'https:///freedombox/login',
        'ftp://example.com',
        'https://[aabb::ccdd',  # Invalid IPv6
    ])
def test_is_safe_url_invalid_url(url):
    """Test invalid URLs for safe URL checks."""
    assert not is_safe_url(url)


def test_json_exception(rf):
    """Test handling exceptions in JSON views."""

    @json_exception
    def error_view(request):
        raise Exception('exception-message')

    assert error_view.__name__ == 'error_view'

    request = rf.get('')
    response = error_view(request)
    assert response.status_code == 500
    json_content = json.loads(response.content)
    assert not json_content['result']
    assert json_content['error_name'] == 'Exception'
    assert json_content['error_string'] == 'exception-message'

    def success_view(request):
        return JsonResponse({'result': True, 'error_string': None})

    request = rf.get('')
    response = success_view(request)
    assert response.status_code == 200
    json_content = json.loads(response.content)
    assert json_content['result']
    assert json_content['error_string'] is None



# SPDX-License-Identifier: AGPL-3.0-or-later
"""URLs for the Miniflux module."""

from django.urls import re_path

from .views import CreateAdminUserView, MinifluxAppView, ResetUserPasswordView

urlpatterns = [
    re_path(r'^apps/miniflux/$', MinifluxAppView.as_view(), name='index'),
    re_path(r'^apps/miniflux/create-admin-user/$',
            CreateAdminUserView.as_view(), name='create-admin-user'),
    re_path(r'^apps/miniflux/reset-user-password/$',
            ResetUserPasswordView.as_view(), name='reset-user-password'),
]



# SPDX-License-Identifier: AGPL-3.0-or-later

from django.utils.translation import gettext_lazy as _

clients = [{
    'name': _('WordPress'),
    'platforms': [{
        'type': 'web',
        'url': '/wordpress/'
    }]
}]

backup = {
    'data': {
        'files': [
            '/var/lib/plinth/backups-data/wordpress-database.sql',
            '/etc/wordpress/is_public'
        ],
        'directories': ['/var/lib/wordpress/']
    },
    'secrets': {
        'directories': ['/etc/wordpress/']
    },
}

tags = [_('Website'), _('Blog'), _('Content management system')]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
FreedomBox app to configure name services.
"""

import json
import logging
import pathlib
import subprocess
from typing import Iterator

from django.utils.translation import gettext_lazy as _
from django.utils.translation import gettext_noop

from plinth import app as app_module
from plinth import cfg, glib, kvstore, menu, network, setup
from plinth.daemon import Daemon
from plinth.diagnostic_check import (DiagnosticCheck,
                                     DiagnosticCheckParameters, Result)
from plinth.modules.backups.components import BackupRestore
from plinth.modules.names.components import DomainType
from plinth.package import Packages
from plinth.privileged import service as service_privileged
from plinth.signals import (domain_added, domain_removed, post_hostname_change,
                            pre_hostname_change)
from plinth.utils import format_lazy

from . import components, manifest, privileged

logger = logging.getLogger(__name__)

_description = [
    format_lazy(
        _('Name Services provides an overview of the ways {box_name} can be '
          'reached from the public Internet: domain name, Tor onion service, '
          'and Pagekite. For each type of name, it is shown whether the HTTP, '
          'HTTPS, and SSH services are enabled or disabled for incoming '
          'connections through the given name.'), box_name=(cfg.box_name))
]


class NamesApp(app_module.App):
    """FreedomBox app for names."""

    app_id = 'names'

    _version = 4

    can_be_disabled = False

    def __init__(self) -> None:
        """Create components for the app."""
        super().__init__()
        info = app_module.Info(app_id=self.app_id, version=self._version,
                               is_essential=True, name=_('Name Services'),
                               icon='fa-tags', description=_description,
                               manual_page='NameServices', tags=manifest.tags)
        self.add(info)

        menu_item = menu.Menu('menu-names', info.name, info.icon, info.tags,
                              'names:index',
                              parent_url_name='system:visibility', order=10)
        self.add(menu_item)

        # 'ip' utility is needed from 'iproute2' package.
        packages = Packages('packages-names', ['iproute2'])
        self.add(packages)

        domain_type = DomainType('domain-type-static', _('Domain (regular)'),
                                 delete_url='names:domain-delete',
                                 add_url='names:domain-add',
                                 can_have_certificate=True, priority=100)
        self.add(domain_type)

        daemon = ResolvedDaemon('daemon-names', 'systemd-resolved')
        self.add(daemon)

        backup_restore = BackupRestore('backup-restore-names',
                                       **manifest.backup)
        self.add(backup_restore)

    @staticmethod
    def post_init():
        """Perform post initialization operations."""
        domain_added.connect(on_domain_added)
        domain_removed.connect(on_domain_removed)

        # Register domain with Name Services module.
        for domain in domains_list() + privileged.get_old_domains():
            domain_added.send_robust(sender='names',
                                     domain_type='domain-type-static',
                                     name=domain, services='__all__')

        # Schedule installation of systemd-resolved if not already installed.
        if not is_resolved_installed():
            # Try to install the package hourly.
            glib.schedule(3600, install_systemd_resolved)

    def diagnose(self) -> list[DiagnosticCheck]:
        """Run diagnostics and return the results."""
        results = super().diagnose()
        if is_resolved_installed():
            results.append(diagnose_resolution('deb.debian.org'))

        return results

    def setup(self, old_version):
        """Install and configure the app."""
        super().setup(old_version)

        if not is_resolved_installed():
            try:
                # Requires internet connectivity and could fail
                privileged.install_resolved()
                privileged.set_resolved_configuration(dns_fallback=True)
            except Exception:
                pass

        if old_version < 4:
            domains = domains_list() + privileged.get_old_domains()
            _domains_set(domains)
            privileged.domain_delete_all()

        if is_resolved_installed():
            # Fresh install or upgrading to version 2
            if old_version < 2:
                privileged.set_resolved_configuration(dns_fallback=True)

            # Load the configuration files for systemd-resolved provided by
            # FreedomBox.
            service_privileged.restart('systemd-resolved')

            # After systemd-resolved is freshly installed, /etc/resolve.conf
            # becomes a symlink to configuration pointing to systemd-resovled
            # stub resolver. However, the old contents are not fed from
            # network-manager (if it was present earlier and wrote to
            # /etc/resolve.conf). Ask network-manager to feed the DNS servers
            # from the connections it has established to systemd-resolved so
            # that using fallback DNS servers is not necessary.
            network.refeed_dns()

        self.enable()


class ResolvedDaemon(Daemon):
    """Perform work only if systemd-resolved is installed."""

    def is_enabled(self):
        """Return if the daemon/unit is enabled."""
        if not is_resolved_installed():
            return True

        return super().is_enabled()

    def enable(self):
        """Run operations to enable the daemon/unit."""
        if is_resolved_installed():
            super().enable()

    def disable(self):
        """Run operations to disable the daemon/unit."""
        if is_resolved_installed():
            super().disable()

    def is_running(self):
        """Return whether the daemon/unit is running."""
        if not is_resolved_installed():
            return True

        return super().is_running()

    def diagnose(self) -> list[DiagnosticCheck]:
        """Check if the daemon is running and listening on expected ports."""
        if not is_resolved_installed():
            return [
                DiagnosticCheck(
                    'names-resolved-installed',
                    gettext_noop('Package systemd-resolved is installed'),
                    Result.WARNING, {}, self.component_id)
            ]

        return super().diagnose()


def domains_list() -> list[str]:
    """Return a list of domains from configuration."""
    return json.loads(kvstore.get_default('domains', '[]'))


def domain_add(domain_name: str):
    """Add a domain to configuration."""
    domains = domains_list()
    if domain_name in domains:
        return

    domains.append(domain_name)
    _domains_set(domains)


def domain_delete(domain_name: str):
    """Remove a domain from configuration."""
    domains = domains_list()
    if domain_name not in domains:
        return

    domains.remove(domain_name)
    _domains_set(domains)


def _domains_set(domains: list[str]):
    """Set the full list of domains in the configuration."""
    kvstore.set('domains', json.dumps(domains))


def install_systemd_resolved(_data):
    """Re-run setup on app to install systemd-resolved."""
    if not is_resolved_installed():
        setup.run_setup_on_app('names', rerun=True)


def diagnose_resolution(domain: str) -> DiagnosticCheck:
    """Perform a diagnostic check for whether a domain can be resolved."""
    result = Result.NOT_DONE
    try:
        subprocess.run(['resolvectl', 'query', '--cache=no', domain],
                       stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL,
                       check=True)
        result = Result.PASSED
    except subprocess.CalledProcessError:
        result = Result.FAILED

    description = gettext_noop('Resolve domain name: {domain}')
    parameters: DiagnosticCheckParameters = {'domain': domain}
    return DiagnosticCheck('names-resolve', description, result, parameters)


def on_domain_added(sender: str, domain_type: str, name: str = '',
                    description: str = '',
                    services: str | list[str] | None = None, **kwargs):
    """Add domain to global list."""
    if not domain_type:
        return

    if not name:
        return
    if not services:
        services = []

    components.DomainName('domain-' + sender + '-' + name, name, domain_type,
                          services)
    logger.info('Added domain %s of type %s with services %s', name,
                domain_type, str(services))

    # HACK: Call from here to here ensure that the on_domain_added can perform
    # DomainName.list() iteration and get an updated list of domains. This
    # won't happen if in the signal calling order, oidc module get notified
    # first. A proper fix is being worked on as a complete overhaul to domain
    # change notification mechanism.
    from plinth.modules import oidc
    oidc.on_domain_added(sender, domain_type, name, description, services,
                         **kwargs)


def on_domain_removed(sender: str, domain_type: str, name: str = '', **kwargs):
    """Remove domain from global list."""
    if name:
        component_id = 'domain-' + sender + '-' + name
        components.DomainName.get(component_id).remove()
        logger.info('Removed domain %s of type %s', name, domain_type)

        # HACK: Call from here to here ensure that the on_domain_remove can
        # perform DomainName.list() iteration and get an updated list of
        # domains. This won't happen if in the signal calling order, oidc
        # module get notified first. A proper fix is being worked on as a
        # complete overhaul to domain change notification mechanism.
        from plinth.modules import oidc
        oidc.on_domain_removed(sender, domain_type, name, **kwargs)
    else:
        for domain_name in components.DomainName.list():
            if domain_name.domain_type.component_id == domain_type:
                domain_name.remove()

                logger.info('Remove domain %s of type %s', domain_name.name,
                            domain_type)


######################################################
# Domain utilities meant to be used by other modules #
######################################################


def get_hostname():
    """Return the hostname."""
    process = subprocess.run(['hostnamectl', 'hostname', '--static'],
                             stdout=subprocess.PIPE, check=True)
    return process.stdout.decode().strip()


def set_hostname(hostname):
    """Set machine hostname and send signals before and after."""
    old_hostname = get_hostname()

    pre_hostname_change.send_robust(sender='names', old_hostname=old_hostname,
                                    new_hostname=hostname)

    logger.info('Changing hostname to - %s', hostname)
    privileged.set_hostname(hostname)

    post_hostname_change.send_robust(sender='names', old_hostname=old_hostname,
                                     new_hostname=hostname)


def get_available_tls_domains() -> Iterator[str]:
    """Return an iterator with all domains able to have a certificate."""
    return (domain.name for domain in components.DomainName.list()
            if domain.domain_type.can_have_certificate)


def is_resolved_installed():
    """Return whether systemd-resolved is installed."""
    return pathlib.Path('/usr/bin/resolvectl').exists()



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
FreedomBox app to configure Transmission server.
"""

from django.urls import reverse_lazy
from django.utils.translation import gettext_lazy as _

from plinth import app as app_module
from plinth import cfg, frontpage, menu
from plinth.config import DropinConfigs
from plinth.daemon import Daemon
from plinth.modules.apache.components import Webserver
from plinth.modules.backups.components import BackupRestore
from plinth.modules.firewall.components import (Firewall,
                                                FirewallLocalProtection)
from plinth.modules.users import add_user_to_share_group
from plinth.modules.users.components import UsersAndGroups
from plinth.package import Packages
from plinth.utils import format_lazy

from . import manifest, privileged

_description = [
    _('Transmission is a BitTorrent client with a web interface.'),
    _('BitTorrent is a peer-to-peer file sharing protocol. '
      'Note that BitTorrent is not anonymous.'),
    _('Please do not change the default port of the Transmission daemon.'),
    format_lazy(
        _('Compared to <a href="{deluge_url}">'
          'Deluge</a>, Transmission is simpler and lightweight but is less '
          'customizable.'), deluge_url=reverse_lazy('deluge:index')),
    format_lazy(
        _('It can be accessed by <a href="{users_url}">any user</a> on '
          '{box_name} belonging to the bit-torrent group.'),
        box_name=_(cfg.box_name), users_url=reverse_lazy('users:index')),
    format_lazy(
        _('In addition to the web interface, mobile and desktop apps can also '
          'be used to remotely control Transmission on {box_name}. To '
          'configure remote control apps, use the URL '
          '<a href="/transmission-remote/rpc">/transmission-remote/rpc</a>.'),
        box_name=_(cfg.box_name)),
    format_lazy(
        _('<a href="{samba_url}">Samba</a> shares can be set as the '
          'default download directory from the dropdown menu below.'),
        samba_url=reverse_lazy('samba:index')),
    format_lazy(
        _('After a download has completed, you can also access your files '
          'using the <a href="{sharing_url}">Sharing</a> app.'),
        sharing_url=reverse_lazy('sharing:index'))
]

SYSTEM_USER = 'debian-transmission'


class TransmissionApp(app_module.App):
    """FreedomBox app for Transmission."""

    app_id = 'transmission'

    _version = 8

    DAEMON = 'transmission-daemon'

    def __init__(self) -> None:
        """Create components for the app."""
        super().__init__()

        groups = {
            'bit-torrent': _('Download files using BitTorrent applications')
        }

        info = app_module.Info(
            app_id=self.app_id, version=self._version, name=_('Transmission'),
            icon_filename='transmission', description=_description,
            manual_page='Transmission', clients=manifest.clients,
            donation_url='https://transmissionbt.com/donate/',
            tags=manifest.tags)
        self.add(info)

        menu_item = menu.Menu('menu-transmission', info.name,
                              info.icon_filename, info.tags,
                              'transmission:index', parent_url_name='apps')
        self.add(menu_item)

        shortcut = frontpage.Shortcut('shortcut-transmission', info.name,
                                      icon=info.icon_filename,
                                      url='/transmission',
                                      clients=info.clients, tags=info.tags,
                                      login_required=True,
                                      allowed_groups=list(groups))
        self.add(shortcut)

        packages = Packages('packages-transmission', ['transmission-daemon'],
                            rerun_setup_on_upgrade=True)
        self.add(packages)

        dropin_configs = DropinConfigs('dropin-configs-transmission', [
            '/etc/apache2/conf-available/transmission-plinth.conf',
        ])
        self.add(dropin_configs)

        firewall = Firewall('firewall-transmission', info.name,
                            ports=['http', 'https',
                                   'transmission-client'], is_external=True)
        self.add(firewall)

        firewall_local_protection = FirewallLocalProtection(
            'firewall-local-protection-transmission', ['9091'])
        self.add(firewall_local_protection)

        webserver = Webserver('webserver-transmission', 'transmission-plinth',
                              urls=['https://{host}/transmission'],
                              last_updated_version=8)
        self.add(webserver)

        daemon = Daemon(
            'daemon-transmission', self.DAEMON, listen_ports=[
                (9091, 'tcp4'),
                (51413, 'tcp4'),
                (51413, 'tcp6'),
                (51413, 'udp4'),
            ])
        self.add(daemon)

        users_and_groups = UsersAndGroups('users-and-groups-transmission',
                                          reserved_usernames=[SYSTEM_USER],
                                          groups=groups)
        self.add(users_and_groups)

        backup_restore = BackupRestore('backup-restore-transmission',
                                       **manifest.backup)
        self.add(backup_restore)

    def setup(self, old_version):
        """Install and configure the app."""
        super().setup(old_version)

        new_configuration = {
            'rpc-whitelist-enabled': False,
            'rpc_whitelist_enabled': False,
            'rpc-authentication-required': False,
            'rpc_authentication_required': False,
        }

        if old_version:
            download_dir = get_download_dir()
            new_configuration['download-dir'] = download_dir
            new_configuration['download_dir'] = download_dir

        privileged.merge_configuration(new_configuration)
        add_user_to_share_group(SYSTEM_USER, TransmissionApp.DAEMON)

        if not old_version:
            self.enable()


def get_download_dir() -> str:
    """Return the configured download directory."""
    configuration = privileged.get_configuration()
    old = configuration.get('download-dir')  # Trixie and older
    if old:
        return old

    return configuration.get('download_dir')  # Forky and newer



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Test network storage.
"""

import pytest

from plinth.modules.backups import store

pytestmark = pytest.mark.django_db

_storages = [
    {
        'path': 'test@nonexistent.org:~/',
        'storage_type': 'ssh',
        'added_by_module': 'test'
    },
    {
        'path': 'test@nonexistent.org:~/tmp/repo/',
        'storage_type': 'ssh',
        'added_by_module': 'test'
    },
]


def test_add():
    """Add a storage item"""
    storage = _storages[0]
    uuid = store.update_or_add(storage)
    _storage = store.get(uuid)
    assert _storage['path'] == storage['path']


def test_add_invalid():
    """Add a storage item"""
    storage_with_missing_type = {
        'path': 'test@nonexistent.org:~/tmp/repo/',
        'added_by_module': 'test'
    }
    with pytest.raises(ValueError):
        store.update_or_add(storage_with_missing_type)


def test_remove():
    """Add and remove storage items"""
    storage = _storages[0]
    uuid = None
    for storage in _storages:
        uuid = store.update_or_add(storage)

    storages = store.get_storages()
    assert len(storages) == 2
    store.delete(uuid)
    storages = store.get_storages()
    assert len(storages) == 1


def test_update():
    """Update existing storage items"""
    uuid = None
    for storage in _storages:
        uuid = store.update_or_add(storage)

    storage = store.get(uuid)
    new_path = 'test@nonexistent.org:~/tmp/repo_new/'
    storage['path'] = new_path
    store.update_or_add(storage)
    _storage = store.get(uuid)
    assert _storage['path'] == new_path



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
URLs for the PageKite module
"""

from django.urls import re_path

from .views import AddCustomServiceView, ConfigurationView, DeleteServiceView

urlpatterns = [
    re_path(r'^sys/pagekite/$', ConfigurationView.as_view(), name='index'),
    re_path(r'^sys/pagekite/services/custom/add/$',
            AddCustomServiceView.as_view(), name='add-custom-service'),
    re_path(r'^sys/pagekite/services/custom/delete/$',
            DeleteServiceView.as_view(), name='delete-custom-service'),
]
